/** 
 * \file BipartiteFrustration.h
 *  
 * \brief Implemention of an algorithm to calculate the
 * Bipartite Edge Frustration of a fullerene graph. Uses
 * (as a subroutine) the BLOSSOM V matching algorithm by Kolmogorov.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date May 2014
 *
 * \attention The following notice does not apply to the BLOSSOM V code
 *            (as it has its own license). This program is distributed
 *            WITHOUT ANY WARRANTY, and it may be freely redistributed
 *            under the condition that the copyright notices  are not
 *            removed,  and no compensation is received. Private, research,
 *            and institutional use is free. Distribution of this  code as
 *            part of a commercial system  is permissible ONLY  BY DIRECT
 *            ARRANGEMENT WITH THE AUTHOR.
 */
 
#ifndef _BIPARTITE_FRUSTRATION_H_
#define _BIPARTITE_FRUSTRATION_H_


//#include <cassert>                          // assert
#include <map>                              // std::map<>
#include <queue>                            // std::queue<>
#include "graphAL.hpp"                      // graph::Graph
#include "PerfectMatching.h"                // BLOSSOM V by Kolmogorov

using namespace std;

// -----------------------------------------------------------------
//
// Type definitions
//
// -----------------------------------------------------------------

/**
 * \typedef Graph
 *
 * \brief Defines Graph as an alias for graph::GraphAM
 */    
typedef graph::Graph Graph;

/**
 * \typedef Graph
 *
 * \brief Defines AdjacencyListIterator as an alias for graph::AdjacencyListIterator
 */    
typedef Graph::AdjacencyListIterator AdjacencyListIterator;

// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

/**
 * \fn int bipartiteEdgeFrustration(Graph& g)
 *
 * \brief Implementation of an algorithm to calculate the Bipartite Edge Frustration
 * of a given fullerene graph (using its dual as input).
 *
 * \param g       The dual of a given fullerene graph.
 * \param level   Array that contains the level of each vertex in the BFS (+1 empty zero position)
 * \param visited Array that helps keep track of which vertices were visited in the BFS (+1 empty zero position)
 *
 * \return The bipartite edge frustration of the given fullerene graph.
 */
int bipartiteEdgeFrustration(Graph& g, unsigned* level, bool* visited) {
    // Data to be used in BLOSSOM V
    int numberOfVerticesK12 = 12;
    int numberOfEdgesK12    = 66;
    int e                   = 0;
    int weights [numberOfEdgesK12];
    int edges   [numberOfEdgesK12 * 2];
    PerfectMatching *pm     = new PerfectMatching(numberOfVerticesK12, numberOfEdgesK12);
    unsigned n              = g.getNumberOfVertices();
    int mapV                = 0;
    map<unsigned, int> mapping; // Maps the index of pentagon vertices in g to the range [0..11]

    // Iterate through all the vertices
    for (unsigned i = 1; i <= n; ++i) {
        // If this vertex corresponds to a pentagon...
        if (g.getDegree(i) == 5) {
            if (mapV < numberOfVerticesK12) {
                mapping[i] = mapV;
                ++mapV;
            }
            // Reset auxiliary variables to default values
            queue<unsigned> bfs;
            for (unsigned j = 0; j <= n; ++j) {
                level[j]    = 0;
                visited[j]  = false;
            }

            // Keeps track of the number of pentagons found
            unsigned countPentagons = 0;

            // Start BFS at i
            bfs.push(i);
            visited[i] = true;
            while (!bfs.empty()) {
                unsigned v = bfs.front();
                bfs.pop();
                if (g.getDegree(v) == 5) {
                    if (i < v) {    // So as to not count twice
                        map<unsigned, int>::iterator it = mapping.find(v);
                        if (it == mapping.end() && mapV < numberOfVerticesK12) {
                            mapping[v] = mapV;
                            ++mapV;
                        }
                        weights[e]       = (int)level[v];
                        edges[2 * e]     = mapping[i];
                        edges[2 * e + 1] = mapping[v];
                        pm->AddEdge(edges[2 * e], edges[2 * e + 1], weights[e]);
                        ++e;
                    }
                    ++countPentagons;
                    if (countPentagons == 12) {
                        break;
                    }
                }
                for (AdjacencyListIterator it = g.getNeighboursBegin(v); it != g.getNeighboursEnd(v); ++it) {
                    unsigned w = (*it);
                    if (!visited[w]) {
                        bfs.push(w);
                        visited[w] = true;
                        level[w]   = level[v] + 1;
                    }
                }
            }
        }
    }

    // Calculate the matching
    struct PerfectMatching::Options options;
    options.verbose = false;
    pm->options = options;
    pm->Solve();
    int cost = -1;
    int res  = CheckPerfectMatchingOptimality(numberOfVerticesK12, numberOfEdgesK12, edges, weights, pm);
    if (res == 0) {
        cost = (int)ComputePerfectMatchingCost(numberOfVerticesK12, numberOfEdgesK12, edges, weights, pm);
    }
    delete pm;
    return cost;
}

#endif // _BIPARTITE_FRUSTRATION_H_