/** 
 * \file FowlerManolopoulos.h
 *  
 * \brief Implemention of a simple algorithm to calculate the
 * Fowler-Manolopoulos criterion of a fullerene graph
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
 
#ifndef _FOWLER_MANOLOPOULOS_H_
#define _FOWLER_MANOLOPOULOS_H_


//#include <cassert>                          // assert
#include <cmath>                           // sqrt

using namespace std;

const unsigned MAX_UNSIGNED = numeric_limits<unsigned int>::max();

// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

/**
 * \fn double fowlerManolopoulosCriterion(unsigned* hexNeighbourhoodIndices)
 *
 * \brief Implementation of a simple algorithm to calculate the Fowler-Manolopoulos
 * criterion of a given fullerene graph (in the form of an array containing hexagon
 * neighbourhood indices, obtained from its dual).
 *
 * \param hexNeighbourhoodIndices An array containin hexagon neighbourhood indices in the form <h0, h1, h2, h3, h4, h5, h6>
 *
 * \return The standard deviation of the neighbourhood indices distribution.
 */
double fowlerManolopoulosCriterion(unsigned* hexNeighbourhoodIndices) {

    unsigned summHk   = 0; // Summation(0..6, Hk)
    unsigned summKHk  = 0; // Summation(0..6, k * Hk)
    unsigned summK2Hk = 0; // Summation(0..6, k^2 * Hk)
    for (unsigned k = 0; k < 7; ++k) {
        summHk   += hexNeighbourhoodIndices[k];
        summKHk  += k * hexNeighbourhoodIndices[k];
        summK2Hk += k * k * hexNeighbourhoodIndices[k];
    }

    double k2 = ((double) summK2Hk) / ((double) summHk);
    double k  = ((double) (summKHk * summKHk) )  / ((double) (summHk * summHk) );
    // k is already squared here

    return sqrt(k2 - k);
    
    
}

#endif // _FOWLER_MANOLOPOULOS_H_
