/** 
 * \file FriesNumber.h
 *  
 * \brief Implemention of an algorithm to calculate the fries number
 *        of a given fullerene graph.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date Aug 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
 
#ifndef _FRIES_NUMBER_H_
#define _FRIES_NUMBER_H_


//#include <cassert>                          // assert
#include <glpk.h>
//#include <scip/scip.h>
//#include <scip/scipdefplugins.h>

using namespace std;

// -----------------------------------------------------------------
//
// Type definitions
//
// -----------------------------------------------------------------

/**
 * \typedef MatrixD
 *
 * \brief Defines Matrix as an alias for Eigen::MatrixXd
 *
typedef Eigen::MatrixXd MatrixD;*/

// Struct to store each ordered pair (u, v) and the associated face (IDs start at 1, 0 is "null")
typedef struct OrderedVertexPair {
    unsigned first  ;
    unsigned second ;
    unsigned face   ;
    unsigned varID  ; // GLPK
    //SCIP_VAR* var ; // SCIP
} OrderedVertexPair ;

// Struct to store face data: an ID and the associated pair
typedef struct Face {
    unsigned           ID      ;
    OrderedVertexPair* pair    ;
} Face;

// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------


/**
 * \fn OrderedVertexPair* symmetric(OrderedVertexPair uv, OrderedVertexPair** a)
 *
 * \brief Returns the symmetric pair, (v, u), of (u, v)
 *
 * \param uv An ordered vertex pair
 * \param a  A 2-dimensional array of ordered vertex pairs
 *
 * \return Pointer to the symmetric pair, (v, u), of (u, v)
 */
OrderedVertexPair* symmetric(OrderedVertexPair uv, OrderedVertexPair** a) {
    unsigned u = uv.first;
    unsigned v = uv.second;
    if (a[v][0].second == u) {
        return &a[v][0];
    }
    if (a[v][1].second == u) {
        return &a[v][1];
    }
    if (a[v][2].second == u) {
        return &a[v][2];
    }
    return 0;
}

/**
 * \fn OrderedVertexPair* nextV(OrderedVertexPair uv, OrderedVertexPair** a)
 *
 * \brief Returns the pair (u, x) corresponding to the vertex, x, following v in the list of neighbours of u
 *
 * \param uv An ordered vertex pair
 * \param a  A 2-dimensional array of ordered vertex pairs
 *
 * \return Pointer to the pair (u, x) corresponding to the vertex, x, following v in the list of neighbours of u
 */
OrderedVertexPair* nextV(OrderedVertexPair uv, OrderedVertexPair** a) {
    unsigned u = uv.first;
    unsigned v = uv.second;
    if (a[u][0].second == v) {
        return &a[u][1];
    }
    if (a[u][1].second == v) {
        return &a[u][2];
    }
    if (a[u][2].second == v) {
        return &a[u][0];
    }
    return 0;
}


/**
 * \fn OrderedVertexPair* nextF(OrderedVertexPair uv, OrderedVertexPair** a)
 *
 * \brief Returns the pair (v, w) corresponding to the next edge around the left face (from u to v) in counter-clockwise order.
 *
 * \param uv An ordered vertex pair
 * \param a  A 2-dimensional array of ordered vertex pairs
 *
 * \return Pointer to the pair (v, w) corresponding to the next edge around the left face (from u to v) in counter-clockwise order.
 */
OrderedVertexPair* nextF(OrderedVertexPair uv, OrderedVertexPair** a) {
    return nextV((*symmetric(uv, a)), a);
}

#endif // _FRIES_NUMBER_H_
