/** 
 * \file edge.hpp
 *  
 * \brief Definition of class Edge, which represents an edge in a graph.
 *
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 3.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef _EDGE_HPP_
#define _EDGE_HPP_

#include <vector>       // std::vector

/** 
* \defgroup GRAPHNameSpace Namespace graph.
* @{
*/

/**
* \namespace graph
*
* \brief   The   namespace   graph   contains   the   definition   and
* implementation of  classes for representing general  graphs.
*
*/

namespace graph {

    /**
     * Forward definition of classes.
     */
    class Vertex ;

    /**
     * \class Edge
     *
     * \brief This class represents an edge of a general graph.
     */
    class Edge {
        public:

            // ---------------------------------------------------------------
            //
            // Type definitions
            //
            // ---------------------------------------------------------------

            /**
             * \typedef Vertex
             *
             * \brief  Defines  Vertex as  an  alias for graph::Vertex.
             */    
            typedef graph::Vertex Vertex ;

            // ---------------------------------------------------------------
            //
            // Public methods.
            //
            // ---------------------------------------------------------------
            /**
             * \fn Edge()
             *
             * \brief Creates an empty instance of this class.
             */
            Edge() {
                _id    = 0 ;
                _v     = 0 ;
                _w     = 0 ;
            }
    
            /**
             * \fn Edge( unsigned id, unsigned level, Vertex* v, Vertex* w )
             *
             * \brief Creates an instance of this class.
             *
             * \param id    The ID of this edge.
             * \param level The level of this edge.
             * \param v     A pointer to the first endpoint of this edge.
             * \param w     A pointer to the second endpoint of this edge.
             */
            Edge(unsigned id, Vertex* v, Vertex* w) {
                _id    = id    ;
                _v     = v     ;
                _w     = w     ;
            }
        
            /**
             * \fn ~Edge()
             *
             * \brief Destroys an instance of this class.
             */
            ~Edge() { }
        
            /**
             * \fn unsigned getID()
             *
             * \brief Returns the ID of this edge.
             *
             * \return The ID of this edge.
             */
            inline unsigned getID() const {
                return _id;
            }
        
            /**
             * \fn Vertex* getV()
             *
             * \brief Returns a pointer to the first endpoin of this edge.
             *
             * \return A pointer to the first endpoint of this edge.
             */
            inline Vertex* getV() const {
                return _v;
            }
        
            /**
             * \fn Vertex* getW()
             *
             * \brief Returns a pointer to the second endpoin of this edge.
             *
             * \return A pointer to the second endpoint of this edge.
             */
            inline Vertex* getW() const {
                return _w;
            }
        
            /**
             * \fn void setID( unsigned id )
             *
             * \brief Assigns a new ID to this edge.
             *
             * \param id The new ID of this edge.
             */
            inline void setID(unsigned id) {
                _id = id;
            }
        
            /**
             * \fn void setV( Vertex* v )
             *
             * \brief Assigns a new first endpoint to this edge.
             *
             * \param v A pointer to the new first endpoint of this edge.
             */
            inline void setV(Vertex* v) {
                _v = v;
            }
        
            /**
             * \fn void setW( Vertex* w )
             *
             * \brief Assigns a new second endpoint to this edge.
             *
             * \param w A pointer to the new second endpoint of this edge.
             */
            inline void setW(Vertex* w) {
                _w = w;
            }
            
        private:
            // ---------------------------------------------------------------
            //
            // Private data members
            //
            // ---------------------------------------------------------------
            unsigned _id         ;  ///< The unique ID of this edge.
            Vertex*  _v          ;  ///< Pointer to the first endpoint of this edge.
            Vertex*  _w          ;  ///< Pointer to the second endpoint of this edge.

    };

}

/** @} */ //end of group class.

#endif // _EDGE_HPP_
