/** 
 * \file graph.hpp
 *  
 * \brief Definition of class Graph, which represents a general graph.
 *
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 3.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
 
#ifndef _GRAPH_HPP_
#define _GRAPH_HPP_

#include "vertex.hpp"       // Vertex
#include "edge.hpp"         // Edge

//#include <cassert>          // assert
#include <vector>           // std::vector
#include <queue>            // std::queue
#include <map>              // std::map

/** 
* \defgroup GRAPHNameSpace Namespace graph.
* @{
*/

/**
* \namespace graph
*
* \brief   The   namespace   graph   contains   the   definition   and
* implementation of  classes for representing general  graphs.
*
*/

namespace graph {

    /**
     * \class Graph
     *
     * \brief This class represents a general graph.
     */
    class Graph {
        public:
            // ---------------------------------------------------------------
            //
            // Type definitions
            //
            // ---------------------------------------------------------------

            /**
             * \typedef Vertex
             *
             * \brief Defines Vertex as an alias for graph::Vertex.
             */    
            typedef graph::Vertex Vertex ;

            /**
             * \typedef Edge
             *
             * \brief  Defines  Edge as  an  alias for graph::Edge.
             */    
            typedef graph::Edge Edge ;

            /**
             * \typedef VertexMap
             *
             * \brief  Defines VertexMap as  an  alias for
             *  std::map< unsigned, Vertex* >.
             */    
            typedef std::map< unsigned, Vertex* > VertexMap ;

            /**
             * \typedef VertexMapIterator
             *
             * \brief Defines an iterator for the VertexMap of this graph.
             */    
            typedef typename std::map< unsigned, Vertex* >::iterator VertexMapIterator ;

            /**
             * \typedef EdgeIterator
             *
             * \brief Defines an iterator for the adjacency list of the vertices.
             */    
            typedef typename std::vector< Edge* >::iterator EdgeIterator ;

            // ---------------------------------------------------------------
            //
            // Public methods.
            //
            // ---------------------------------------------------------------
            /**
             * \fn Graph()
             *
             * \brief Creates an empty instance of this class.
             */
            Graph() { }
        
            /**
             * \fn ~Graph()
             *
             * \brief Destroys an instance of this class.
             */
            ~Graph() {
                clear();
            }
        
            /**
             * \fn Vertex* getVertex(unsigned id)
             *
             * \brief Returns the vertex with ID = id, if it is on the graph.
             *
             * \param id ID of the vertex to be returned.
             *
             * \return Vertex with ID = id, or null if it's not on the graph.
             */
            inline Vertex* getVertex(unsigned id) {
                //assert(_vertices.find(id) != _vertices.end());
                //Vertex* v = _vertices[id];
                return _vertices[id];
            }
        
            /**
             * \fn Edge* getEdge(unsigned id)
             *
             * \brief Returns the edge with ID = id, if it's on the graph.
             *
             * \param id ID of the edge to be returned.
             *
             * \return Edge with ID = id, or null if it's not on the graph.
             */
            inline Edge* getEdge(unsigned id) const {
                for (VertexMapIterator it = _vertices.begin(); it != _vertices.end(); ++it) {
                    Vertex* v = (*it).second;
                    //assert(v != 0);
                    for (EdgeIterator eIt = v->incidentEdgesBegin(); eIt != v->incidentEdgesEnd(); ++eIt) {
                        Edge* e = (*eIt);
                        if (e->getID() == id) {
                            return e;
                        }
                    }
                }
                return 0;
            }

            /**
             * \fn bool isEdge(unsigned v, unsigned w)
             *
             * \brief Returns true if there is an edge between the vertices with IDs v and w, and false otherwise
             *
             * \param v ID of the first vertex to be tested
             * \param w ID of the second vertex to be tested
             *
             * \return true if there is an edge between v and w, false otherwise.
             */
            inline bool isEdge(unsigned v, unsigned w) const {
                Vertex* vp = _vertices[v];
                Vertex* wp = _vertices[w];
                return vp->isNeighbour(wp);
            }
        
            /**
             * \fn inline unsigned size()
             *
             * \brief Returns the number of vertices in this graph.
             *
             * \return The number of vertices in this graph.
             */
            inline unsigned size() const {
                return _vertices.size();
            }
        
            /**
             * \fn inline VertexMapIterator verticesBegin()
             *
             * \brief Returns  a vertex iterator set  to the first vertex of the
             * vertex map of this graph.
             *
             * \return A vertex iterator set to the first vertex of the vertex map of this graph.
             */
            inline VertexMapIterator verticesBegin() {
                return _vertices.begin();
            }
        
            /**
             * \fn inline VertexMapIterator verticesEnd()
             *
             * \brief Returns  a vertex iterator set  to the end of the
             * vertex map of this graph.
             *
             * \return A vertex iterator set to the end of the vertex map of this graph.
             */
            inline VertexMapIterator verticesEnd() {
                return _vertices.end();
            }
        
            /**
             * \fn void insertVertex( unsigned id )
             *
             * \brief Adds a new vertex to the graph, if it's not alredy there.
             *
             * \param id The id of the new vertex.
             *
             * \return The newly created vertex.
             */
            inline Vertex* insertVertex(unsigned id) {
                // Only insert the new vertex if it's not alredy in the graph.
                //assert(_vertices.find(id) == _vertices.end());
                
                // Insert it into the graph.
                Vertex* v = new Vertex(id);
                _vertices[id] = v;

                return v;
            }
        
            /**
             * \fn void removeVertex( Vertex* x )
             *
             * \brief Removes the vertex x from the graph, if it's there.
             *
             * \param x The vertex to be removed.
             */
            inline void removeVertex(Vertex* x) {
        
                // Only remove the vertex if it is in the graph.
                //assert(x != 0);
                //assert(_vertices.find(x->getID()) != _vertices.end());

                unsigned id = x->getID();
        
                // Remove all edges incident to this vertex from the graph
                // (from the adjacency list on its neighbours).
                // Loop through all edges incident to this vertex.
                //int count = 0;
                for (EdgeIterator it = x->incidentEdgesBegin(); it != x->incidentEdgesEnd(); ++it) {
                    //++count;
                    Edge*   e = (*it);
                    //assert(e != 0);
                    Vertex* v = e->getV();
                    Vertex* w = e->getW();
                    // Removing each edge from its neighbours adjacency list
                    // (Careful to avoid loops here, or else the removal of edge
                    // e from the adjacency list could invalidate the iterator).
                    if (v->getID() == id && w->getID() != id) {
                        w->removeIncidentEdge(e);
                    }
                    if (v->getID() != id && w->getID() == id) {
                        v->removeIncidentEdge(e);
                    }
                    // Now the only references to it are the e pointer and the
                    // pointer on the adjacency list of this vertex.
                    delete e;
                    // The dangling pointer on the adjacency list will be erased
                    // once x is deleted, just outside of this loop
                }
                //assert(count == 0);
                // Now remove the vertex from the VertexMap.
                _vertices.erase(id);
                // And delete it.
                delete x;
            }
        
            /**
             * \fn void insertEdge( unsigned id, Vertex* v, Vertex* w )
             *
             * \brief Adds a new edge to the graph, if it's not alredy there.
             *
             * \param id The id of the new edge.
             * \param v  A pointer to the first endpoint of the new edge.
             * \param w  A pointer to the second endpoint of the new edge.
             *
             * \return The newly created edge.
             */
            inline Edge* insertEdge(unsigned id, Vertex* v, Vertex* w) {
                // Only insert the new edge if it's linking 2 vertices that are in the graph.
                // We allow edges of multiplicity > 1 and tell edges apart by an unsinged index.
                // So instead of checking if there is alredy another edge linking those vertices,
                // we check if the edge with given index alredy exists.
        
                /*assert(v != 0 && w != 0);
                assert(_vertices.find(v->getID()) != _vertices.end());
                assert(_vertices.find(w->getID()) != _vertices.end());
        
                bool found = false;
                
                for (VertexMapIterator it = _vertices.begin(); it != _vertices.end(); ++it) {
                    Vertex* x = (*it).second;
                    assert(x != 0);
                    for (EdgeIterator eIt = x->incidentEdgesBegin(); eIt != x->incidentEdgesEnd(); ++eIt) {
                        if ((*eIt)->getID() == id) {
                            found = true;
                            break;
                        }
                    }
                    if (found) {
                        break;
                    }
                }
                
                assert(!found);*/
            
                Edge* e = new Edge(id, v, w);
                v->addIncidentEdge(e);
                w->addIncidentEdge(e);

                return e;
            }
        
            /**
             * \fn void removeEdge( Edge* e )
             *
             * \brief Removes the edge e from the graph, if it's there.
             *
             * \param e The edge to be removed.
             */
            inline void removeEdge(Edge* e) {

                Vertex* v = e->getV() ;
                Vertex* w = e->getW() ;
        
                /*bool found = false;

                for (EdgeIterator it = v->incidentEdgesBegin(); it != v->incidentEdgesEnd(); ++it) {
                    if ((*it)->getID() == e->getID()) {
                        found = true;
                        break;
                    }
                }
                
                assert(found);*/

                // Remove e from the adjacency list on V and W.
                v->removeIncidentEdge(e);
                w->removeIncidentEdge(e);
        
                // Finally, delete it (this is the last pointer to it).
                delete e;

            }

            /**
             * \fn void removeEdge( unsigned id, Vertex* v, Vertex* w )
             *
             * \brief Removes the edge (v, w) with ID = id from the graph, if it's there.
             *
             * \param id The id of the edge to be removed.
             * \param v  The first endpoint of the edge to be removed.
             * \param w  The second endpoint of the edge to be removed.
             */
            inline void removeEdge(unsigned id, Vertex* v, Vertex* w) {
        
                /*bool found = false;
                
                for (EdgeIterator it = v->incidentEdgesBegin(); it != v->incidentEdgesEnd(); ++it) {
                    if ((*it)->getID() == id) {
                        found = true;
                        break;
                    }
                }
                
                assert(found);*/

                Edge* e = v->getIncidentEdge(id);
        
                // Remove e from the adjacency list on V and W.
                v->removeIncidentEdge(e);
                w->removeIncidentEdge(e);
        
                // Finally, delete it (this is the last pointer to it).
                delete e;
            
            }
        
            /**
             * \fn void clear()
             *
             * \brief Removes all vertices and edges from the graph.
             */
            void clear() {
                // For each vertex:
                for (VertexMapIterator it = _vertices.begin(); it != _vertices.end(); ++it) {
                    Vertex* x = (*it).second;
                    //assert(x != 0);
                    // For each edge incident on this vertex:
                    for (EdgeIterator it = x->incidentEdgesBegin(); it != x->incidentEdgesEnd(); ++it) {
                        // Remove the reference to it from the adjacency list
                        // of the neighbouring vertex.
                        Edge* e   = (*it);
                        Vertex* v = e->getV();
                        Vertex* w = e->getW();
                        if (v->getID() == x->getID() && w->getID() != x->getID()) {
                            w->removeIncidentEdge(e);
                        }
                        if (v->getID() != x->getID() && w->getID() == x->getID()) {
                            v->removeIncidentEdge(e);
                        }
                        // Now the only references to it are the e pointer and the
                        // pointer on the adjacency list of the X vertex.
                        delete e;
                        // The dangling pointer on the adjacency list will be erased
                        // once x is deleted, just outside of this loop
                    }
                    delete x;
                }
                _vertices.clear();
            }
    
    
        private:
            // ---------------------------------------------------------------
            //
            // Private methods
            //
            // ---------------------------------------------------------------

            
            // ---------------------------------------------------------------
            //
            // Private data members
            //
            // ---------------------------------------------------------------
            VertexMap _vertices   ;  ///< Mapping of vertices IDs to vertices pointers, contains all vertices of this graph.
    };

}

/** @} */ //end of group class.

#endif // _GRAPH_HPP_
