/** 
 * \file graphAL.hpp
 *  
 * \brief Definition of class Graph, which represents a general graph
 *        using an adjacency list. Vertices' IDs range from 1 to N.
 *
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 3.0
 * \date May 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
 
#ifndef _GRAPHAL_HPP_
#define _GRAPHAL_HPP_

//#include <cassert>          // assert
#include <list>             // std::list<>

/** 
* \defgroup GRAPHNameSpace Namespace graph.
* @{
*/

/**
* \namespace graph
*
* \brief   The   namespace   graph   contains   the   definition   and
* implementation of  classes for representing general  graphs.
*
*/

namespace graph {

    /**
     * \class Graph
     *
     * \brief This class represents a general graph using an adjacency list.
     */
    class Graph {
        public:

            // ---------------------------------------------------------------
            //
            // Type definitions
            //
            // ---------------------------------------------------------------

            /**
             * \typedef AdjacencyList
             *
             * \brief   Defines AdjacencyList as an alias for std::list<unsigned>
             */
            typedef std::list<unsigned> AdjacencyList;

            /**
             * \typedef AdjacencyListIterator
             *
             * \brief   Defines AdjacencyListIterator as an alias for AdjacencyList::iterator
             */
            typedef AdjacencyList::iterator AdjacencyListIterator;

            // ---------------------------------------------------------------
            //
            // Public methods.
            //
            // ---------------------------------------------------------------
            /**
             * \fn Graph(unsigned n)
             *
             * \brief Creates an instance of this class with n + 1 vertices and no edges.
             */
            Graph(unsigned n) {
                _numberOfVertices = n + 1;
                _numberOfEdges    = 0;
                _vertices         = new AdjacencyList[_numberOfVertices];
            }
        
            /**
             * \fn ~Graph()
             *
             * \brief Destroys an instance of this class.
             */
            ~Graph() {
                clear();
                delete[] _vertices;
            }
        
            /**
             * \fn bool isEdge(unsigned v, unsigned w)
             *
             * \brief Returns true if there is an edge between v and w, and false otherwise.
             *
             * \param v ID of the first vertex to be tested.
             * \param w ID of the second vertex to be tested.
             *
             * \return True if there is an edge between v and w, and false otherwise.
             */
            inline bool isEdge(unsigned v, unsigned w) const {
                //assert(v < _numberOfVertices && w < _numberOfVertices);
                for (AdjacencyList::iterator it = _vertices[v].begin(); it != _vertices[v].end(); ++it) {
                    if ((*it) == w) {
                        return true;
                    }
                }
                return false;
            }
        
            /**
             * \fn unsigned getDegree(unsigned v)
             *
             * \brief Returns the degree of vertex v, assuming it's on the graph.
             *
             * \param v ID of the vertex to be checked.
             *
             * \return The degree of vertex v, assuming it's on the graph.
             */
            inline unsigned getDegree(unsigned v) const {
                //assert(v < _numberOfVertices);
                return _vertices[v].size();
            }

            /**
             * \fn AdjacencyListIterator getNeighboursBegin(unsigned v)
             *
             * \brief Returns an iterator to the beginning of the list with the neighbours of vertex v
             *
             * \param v ID of the vertex to be checked.
             *
             * \return An iterator to the beginning of the list with the neighbours of vertex v.
             */
            inline AdjacencyListIterator getNeighboursBegin(unsigned v) {
                //assert(v < _numberOfVertices);
                return _vertices[v].begin();
            }

            /**
             * \fn AdjacencyListIterator getNeighboursEnd(unsigned v)
             *
             * \brief Returns an iterator to the end of the list with the neighbours of vertex v
             *
             * \param v ID of the vertex to be checked.
             *
             * \return An iterator to the end of the list with the neighbours of vertex v.
             */
            inline AdjacencyListIterator getNeighboursEnd(unsigned v) {
                //assert(v < _numberOfVertices);
                return _vertices[v].end();
            }
        
            /**
             * \fn inline unsigned getNumberOfVertices()
             *
             * \brief Returns the number of vertices in this graph.
             *
             * \return The number of vertices in this graph.
             */
            inline unsigned getNumberOfVertices() const {
                return _numberOfVertices - 1;
            }

            /**
             * \fn inline unsigned getNumberOfEdges()
             *
             * \brief Returns the number of edges in this graph.
             *
             * \return The number of edges in this graph.
             */
            inline unsigned getNumberOfEdges() const {
                return _numberOfEdges;
            }
        
            /**
             * \fn void insertEdge(unsigned v, unsigned w)
             *
             * \brief Adds a new edge to the graph, assuming it's not alredy there.
             *
             * \param v       ID of the first vertex of the edge.
             * \param w       ID of the second vertex of the edge.
             */
            inline void insertEdge(unsigned v, unsigned w) {
                //assert(v < _numberOfVertices && w < _numberOfVertices);
                _vertices[v].push_back(w);
                _vertices[w].push_back(v);
                _numberOfEdges++;
            }
        
            /**
             * \fn void clear()
             *
             * \brief Removes all edges from the graph.
             */
            void clear() {
                for (unsigned i = 0; i < _numberOfVertices; ++i) {
                    _vertices[i].clear();
                }
                _numberOfEdges = 0;
            }
    
    
        private:
            // ---------------------------------------------------------------
            //
            // Private methods
            //
            // ---------------------------------------------------------------

            
            // ---------------------------------------------------------------
            //
            // Private data members
            //
            // ---------------------------------------------------------------
            unsigned       _numberOfVertices ;
            unsigned       _numberOfEdges    ;
            AdjacencyList* _vertices         ;
    };

}

/** @} */ //end of group class.

#endif // _GRAPHAL_HPP_
