/** 
 * \file graphAM.hpp
 *  
 * \brief Definition of class Graph, which represents a general graph
 *        using an adjacency matrix (with weights on the edges and an
 *        aditional array to store degrees separately).
 *        Vertices' IDs range from 1 to N.
 *
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 3.0
 * \date May 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
 
#ifndef _GRAPHAM_HPP_
#define _GRAPHAM_HPP_

//#include <cassert>          // assert

/** 
* \defgroup GRAPHNameSpace Namespace graph.
* @{
*/

/**
* \namespace graph
*
* \brief   The   namespace   graph   contains   the   definition   and
* implementation of  classes for representing general  graphs.
*
*/

namespace graph {

    /**
     * \class Graph
     *
     * \brief This class represents a general graph using an adjacency matrix
     *        (with weights on the edges and an aditional array to store degrees
     *         separately).
     */
    class Graph {
        public:

            // ---------------------------------------------------------------
            //
            // Public methods.
            //
            // ---------------------------------------------------------------
            /**
             * \fn Graph(unsigned n)
             *
             * \brief Creates an instance of this class with n + 1 vertices and no edges.
             */
            Graph(unsigned n) {
                _numberOfVertices = n + 1                            ;
                _numberOfEdges    = 0                                ;
                _degree           = new unsigned[_numberOfVertices] ;
                _vertices         = new int*[_numberOfVertices]      ;
                for (unsigned i = 0; i < _numberOfVertices; ++i) {
                    _degree[i]    = 0;
                    _vertices[i]  = new int[_numberOfVertices];
                    for (unsigned j = 0; j < _numberOfVertices; ++j) {
                        _vertices[i][j] = 0;
                    }
                }
            }
        
            /**
             * \fn ~Graph()
             *
             * \brief Destroys an instance of this class.
             */
            ~Graph() {
                delete[] _degree;
                for (unsigned i = 0; i < _numberOfVertices; ++i) {
                    delete[] _vertices[i];
                }
                delete[] _vertices;
            }
        
            /**
             * \fn bool isEdge(unsigned v, unsigned w)
             *
             * \brief Returns true if there is an edge between v and w, and false otherwise.
             *
             * \param v ID of the first vertex to be tested.
             * \param w ID of the second vertex to be tested.
             *
             * \return True if there is an edge between v and w, and false otherwise.
             */
            inline bool isEdge(unsigned v, unsigned w) const {
                //assert(v < _numberOfVertices && w < _numberOfVertices);
                return _vertices[v][w] != 0;
            }
        
            /**
             * \fn unsigned getDegree(unsigned v)
             *
             * \brief Returns the degree of vertex v, assuming it's on the graph.
             *
             * \param v ID of the vertex to be checked.
             *
             * \return The degree of vertex v, assuming it's on the graph.
             */
            inline unsigned getDegree(unsigned v) const {
                //assert(v < _numberOfVertices);
                /* Without using _degree
                unsigned degree = 0;
                for (unsigned j = 0; j < _numberOfVertices; ++j) {
                    if (_vertices[v][j] != 0) {
                        ++degree;
                    }
                }
                return degree;
                */
                return _degree[v];
            }
        
            /**
             * \fn inline unsigned getNumberOfVertices()
             *
             * \brief Returns the number of vertices in this graph.
             *
             * \return The number of vertices in this graph.
             */
            inline unsigned getNumberOfVertices() const {
                return _numberOfVertices - 1;
            }

            /**
             * \fn inline unsigned getNumberOfEdges()
             *
             * \brief Returns the number of edges in this graph.
             *
             * \return The number of edges in this graph.
             */
            inline unsigned getNumberOfEdges() const {
                return _numberOfEdges;
            }
        
            /**
             * \fn void insertEdge(unsigned v, unsigned w, int weight = 1)
             *
             * \brief Adds a new edge to the graph, assuming it's not alredy there.
             *
             * \param v       ID of the first vertex of the edge.
             * \param w       ID of the second vertex of the edge.
             * \param weight  Optional weight of the new edge. Must not be zero.
             */
            inline void insertEdge(unsigned v, unsigned w, int weight = 1) {
                //assert(v < _numberOfVertices && w < _numberOfVertices);
                //assert(_vertices[v][w] == 0);
                //assert(weight != 0);
                _vertices[v][w] = weight;
                _numberOfEdges++;
                _degree[v]++;
                _degree[w]++;
            }
        
            /**
             * \fn void removeEdge(unsigned v, unsigned w)
             *
             * \brief Removes the between vertices v and w from the graph, assuming it's there.
             *
             * \param v ID of the first vertex of the edge.
             * \param w ID of the second vertex of the edge.
             */
            inline void removeEdge(unsigned v, unsigned w) {
                //assert(v < _numberOfVertices && w < _numberOfVertices);
                //assert(_vertices[v][w] != 0);
                //assert(_numberOfEdges > 0);
                //assert(_degree[v] > 0);
                //assert(_degree[w] > 0);
                _vertices[v][w] = 0;
                _numberOfEdges--;
                _degree[v]--;
                _degree[w]--;
            }
        
            /**
             * \fn void clear()
             *
             * \brief Removes all edges from the graph.
             */
            void clear() {
                for (unsigned i = 0; i < _numberOfVertices; ++i) {
                    _degree[i] = 0;
                    for (unsigned j = 0; j < _numberOfVertices; ++j) {
                        _vertices[i][j] = 0;
                    }
                }
                _numberOfEdges = 0;
            }
    
    
        private:
            // ---------------------------------------------------------------
            //
            // Private methods
            //
            // ---------------------------------------------------------------

            
            // ---------------------------------------------------------------
            //
            // Private data members
            //
            // ---------------------------------------------------------------
            unsigned  _numberOfVertices ;
            unsigned  _numberOfEdges    ;
            unsigned* _degree           ;  ///< Stores the degrees of the vertices separately
            int**     _vertices         ;  ///< Adjacency matrix
    };

}

/** @} */ //end of group class.

#endif // _GRAPHAM_HPP_
