/** 
 * \file vertex.hpp
 *  
 * \brief Definition of class Vertex,  which represents a node in a graph.
 *
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 3.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#ifndef _VERTEX_HPP_
#define _VERTEX_HPP_

#include "edge.hpp"     // Edge

//#include <cassert>      // assert
#include <vector>       // std::vector

/** 
* \defgroup GRAPHNameSpace Namespace graph.
* @{
*/

/**
* \namespace graph
*
* \brief   The   namespace   graph   contains   the   definition   and
* implementation of  classes for representing general  graphs.
*
*/

namespace graph {

    /**
     * Forward definition of classes.
     */
    class Edge ;


    /**
     * \class Vertex
     *
     * \brief This class represents a vertex of a general graph.
     */
    class Vertex {
        public:
            // ---------------------------------------------------------------
            //
            // Type definitions
            //
            // ---------------------------------------------------------------

            /**
             * \typedef Edge
             *
             * \brief  Defines  Edge as  an  alias for graph::Edge.
             */    
            typedef graph::Edge Edge ;

            /**
             * \typedef EdgeIterator
             *
             * \brief Defines an iterator for the adjacency list of this vertex.
             */    
            typedef typename std::vector< Edge* >::iterator EdgeIterator ;

            // ---------------------------------------------------------------
            //
            // Public methods.
            //
            // ---------------------------------------------------------------
            /**
             * \fn Vertex()
             *
             * \brief Creates an empty instance of this class.
             */
            Vertex() {
                _id = 0;
            }
        
            /**
             * \fn Vertex( unsigned id )
             *
             * \brief Creates an instance of this class.
             *
             * \param id    The ID of this vertex.
             */
            Vertex(unsigned id) {
                _id = id;
                _incidentEdges.reserve(3);
            }
        
            /**
             * \fn ~Vertex()
             *
             * \brief Destroys an instance of this class.
             */
            ~Vertex() {
                _incidentEdges.clear();
            }
            
            /**
             * \fn unsigned getID()
             *
             * \brief Returns the ID of this vertex.
             *
             * \return The ID of this vertex.
             */
            inline unsigned getID() const {
                return _id;
            }
        
            /**
             * \fn unsigned getDegree()
             *
             * \brief Returns the degree of this vertex.
             *
             * \return The degree of this vertex.
             */
            inline unsigned getDegree() {
                return _incidentEdges.size();
            }

            /**
             * \fn bool isNeighbour(Vertex* w)
             *
             * \brief Returns true if w is a neighbour of this vertex, and false otherwise.
             *
             * \param w Pointer to the supposed neighbour of this vertex.
             *
             * \return True if w is a neighbour of this vertex, false otherwise.
             */
            inline bool isNeighbour(Vertex* w) {
                for (EdgeIterator it = _incidentEdges.begin(); it != _incidentEdges.end(); ++it) {
                    Edge* e = (*it);
                    if ((e->getV() == v && e->getW() == w) || (e->getV() == w && e->getW() == v)) {
                        return true;
                    }
                }
                return false;
            }

            /**
             * \fn Edge* getIncidentEdge( unsigned id )
             *
             * \brief Returns a pointer to the edge incident on this vertex with ID = id.
             *
             * \param id ID of the desired edge.
             *
             * \return A pointer to the edge with ID = id that is incident on this vertex.
             */
            inline Edge* getIncidentEdge(unsigned id) {
                for (EdgeIterator it = _incidentEdges.begin(); it != _incidentEdges.end(); ++it) {
                    if ((*it)->getID() == id) {
                        return (*it);
                    }
                }
                return 0;
            }
        
            /**
             * \fn inline EdgeIterator incidentEdgeBegin()
             *
             * \brief Returns  an edge iterator set  to the first  edge of the
             * adjacency list of this vertex.
             *
             * \return An edge iterator set to the first edge of the adjacency list of this vertex.
             */
            inline EdgeIterator incidentEdgesBegin() {
                return _incidentEdges.begin();
            }
        
            /**
             * \fn inline EdgeIterator incidentEdgeEnd()
             *
             * \brief Returns  an edge iterator set  to the end of the
             * adjacency list of this vertex.
             *
             * \return An edge iterator set to the end of the adjacency list of this vertex.
             */
            inline EdgeIterator incidentEdgesEnd() {
                return _incidentEdges.end();
            }
        
            /**
             * \fn void setID( unsigned id )
             *
             * \brief Assigns a new ID to this vertex.
             *
             * \param id The new ID of this vertex.
             */
            inline void setID(unsigned id) {
                _id = id;
            }
        
            /**
             * \fn void addIncidentEdge( Edge* e )
             *
             * \brief Adds a new edge to the adjacency list of this vertex, if it's not alredy there.
             *
             * \param e A pointer to the new edge to be added.
             */
            inline void addIncidentEdge(Edge* e) {
                /*unsigned eID    = e->getID();
                bool     foundE = false     ;
                for (EdgeIterator it = _incidentEdges.begin(); it != _incidentEdges.end(); ++it) {
                    if ((*it)->getID() == eID) {
                        foundE = true;
                        break;
                    }
                }
                assert(!foundE);*/
                _incidentEdges.push_back(e);
            }
        
            /**
             * \fn void removeIncidentEdge( Edge* e )
             *
             * \brief Removes an edge from the adjacency list of this vertex, if it's there.
             *
             * \param e A pointer to the edge to be removed.
             */
            inline void removeIncidentEdge(Edge* e) {
                unsigned eID = e->getID();
                //bool foundE = false;
                for (EdgeIterator it = _incidentEdges.begin(); it != _incidentEdges.end(); ++it) {
                    if ((*it)->getID() == eID) {
                        _incidentEdges.erase(it);
                        //foundE = true;
                        break;
                    }
                }
                //assert(foundE);
            }
            
    
        private:
            // ---------------------------------------------------------------
            //
            // Private data members
            //
            // ---------------------------------------------------------------
            unsigned               _id           ;  ///< The unique ID of this vertex.
            std::vector< Edge* >   _incidentEdges;  ///< List of pointers to the edges incident to this vertex.
    
    };

}

/** @} */ //end of group class.

#endif // _VERTEX_HPP_
