/** 
 * \file testIndependenceNumber.cpp
 *  
 * \brief A driver to test the implementation of the algorithm to calculate
 *        the Independence Number of a fullerene graph.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date May 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
#include <iostream>                         // std::cout, std::endl
#include <fstream>                          // ifstream, ofstream
#include <sstream>                          // ostringstream
#include <string>                           // std::string
#include <cstring>                          // strcmp
#include <cstdlib>                          // exit, EXIT_SUCCESS, EXIT_FAILURE
#include <cstdio>                           // sprintf
#include <ctime>                            // time, clock, CLOCKS_PER_SEC, clock_t 
#include <iomanip>                          // std:setprecision
#include <limits>                           // std::numeric_limits
#include <ios>                              // std::fixed
//#include <cassert>                          // assert
#include <algorithm>                        // std:sort
#include <vector>                           // std::vector
#include <queue>                            // std::priority_queue

#include "IndependenceNumber.h"             // independenceNumber(g)


// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

// Error flag
bool errorDetected = false;

// Struct to store each entry in Top 100 / Bottom 100 table and Comparison class
typedef struct TopBottomData {
    unsigned isomerID ;
    int      value    ;
} TopBottomData;

class TopBottomComparison {
    private:
        bool _reverse;

    public:
        TopBottomComparison(const bool& reverse = false) {
            _reverse = reverse;
        }

        bool operator() (const TopBottomData& lhs, const TopBottomData& rhs) const {
            if (_reverse) {
                return (lhs.value > rhs.value);
            } else {
                return (lhs.value < rhs.value);
            }
        }
};

/**
 * \fn inline bool isUnsigned(const string& s)
 *
 * \brief A function to test if a given string represents an unsigned integer number.
 *
 * \param s The string to be tested.
 *
 * \return True if the string represents an unsigned integer number, false otherwise.
 */
inline bool isUnsignedInteger(const string& s) {
    if (s.empty() || ((!isdigit(s[0])) && (s[0] != '+'))) {
        return false;
    }
    char* p;
    strtoul(s.c_str(), &p, 10);
    return (*p == 0);
}

/**
 * \fn void errorLog(const string& m)
 *
 * \brief A function to log an error message.
 *
 * \param m The message to be logged.
 */
void errorLog(const string& m) {
    // Set flag
    errorDetected = true;

    // Write to stdout
    cout << m << endl;

    // And to log file, with current time
    ofstream errorFile("../data/out/error/Error_Log.txt", ios_base::app);
    time_t now = time(0);
    struct tm  tstruct = *localtime(&now);
    char   buf[80];
    strftime(buf, sizeof(buf), "%d-%m-%Y.%X", &tstruct);
    errorFile << buf << ": " << m << endl;
    errorFile.close();
}

/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief A simple test of the implementation of the algorithm to calculate
 *        the Fowler-Manolopoulos criterion of a fullerene graph, using its
 *        dual as input.
 *
 * \param argc The number of command-line arguments.
 * \param argv The command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {

    //
    // Input validation
    //
    bool      isValidInput = false ;
    unsigned  nStart               ;
    unsigned  nFinish              ;
    string    specialFileName = "" ;

    if (argc == 3) {
        string arg1(argv[1]);
        string arg2(argv[2]);
        if (isUnsignedInteger(arg1) && isUnsignedInteger(arg2)) {
            nStart  = strtoul(argv[1], NULL, 10);
            nFinish = strtoul(argv[2], NULL, 10);
            if (nStart  >= 20     && nStart  != 22 && nStart  % 2 == 0 &&
                nFinish >= nStart && nFinish != 22 && nFinish % 2 == 0) {
                isValidInput = true;
            }
            if (nStart == nFinish) {
                specialFileName = arg1;
            }
        }
    }

    if (!isValidInput) {
        cout << endl
             << "\t Usage: testIndependenceNumber < arg1 >  < arg2 >" 
             << endl
             << "\t\t arg1   n at which to start calculating (an even integer >= 20 and != 22)"
             << endl
             << "\t\t arg2   n at which to stop calculating (an even integer >= arg1 and != 22)"
	         << endl
             << endl;

        return EXIT_FAILURE;
    }

    // Create a file to store the top 100 and bottom 100 isomers of each n...
    string top100Bottom100FileName = "../data/out/all-invariants/Independence_Number_" + specialFileName + ".csv";
    ofstream top100Bottom100File(top100Bottom100FileName.c_str());
    if (!top100Bottom100File) {
        // Or log an error and exit the program.
        errorLog("Unable to create file: all-invariants/Independence_Number_" + specialFileName + ".csv");
        return EXIT_FAILURE;
    }

    // And print the headers
    top100Bottom100File << "value_of_n,top_100_decreasing,bottom_100_increasing,top_100_by_id,bottom_100_by_id" << endl;

    // Now define 2 priority_queue objects (top and bottom 100).
    typedef priority_queue< TopBottomData, vector<TopBottomData>, TopBottomComparison > TopBottomQueue;
    TopBottomQueue top100(TopBottomComparison(true)); // The top 100 queues are inverse priority queues
    TopBottomQueue bottom100; // The bottom 100 queues are normal priority queues
    const unsigned MAX_QUEUE_SIZE = 100;

    //
    // For each value of N in the designated range...
    //
    for (unsigned nCurrent = nStart; nCurrent <= nFinish; nCurrent += 2) {

        // Skip if N is 22, as there are no fullerene graphs with 22 vertices
        if (nCurrent == 22) {
            continue;
        }

        ostringstream convert;
        convert << nCurrent;
        string nString = convert.str();
        if (nCurrent >= 132) {
            nString += "_ipr";
        }

        // File processing timers
        clock_t fileStart;
        clock_t fileEnd;
        cout << "Starting to process file Fullerenes_" << nCurrent << "..." << endl;
        fileStart = clock();

        // Open the corresponding input file
        string inputFileName = "../data/in/Fullerenes_" + nString;
        ifstream inputFile(inputFileName.c_str(), ifstream::binary);
        if (!inputFile) {
            // Or log an error and try to continue if unable to open
            errorLog("Unable to open file: " + inputFileName);
            continue;
        }

        // Create an output file (.csv format)...
        string outputFileName = "../data/out/independence-number/Independence_Number_" + nString + ".csv";
        ofstream outputFile(outputFileName.c_str());
        if (!outputFile) {
            // Or log an error and try to continue if unable to create.
            errorLog("Unable to create file: " + outputFileName);
            inputFile.close();
            continue;
        }

        // And print the headers of the data we will write.
        outputFile << "isomer_id,independence_number" << endl;

        // Read the first 15 characters on the file and check if it
        // really is planar code
        char* planarCodeHeader = new char[15];
        inputFile.get(planarCodeHeader, 16); // one more than buffer size for get to work
        if (strcmp(planarCodeHeader, ">>planar_code<<") != 0) {
            // Or log an error and try to continue if the header is wrong
            errorLog("File is missing planar code header: " + inputFileName + "    Header: " + planarCodeHeader);
            inputFile.close();
            outputFile.close();
            delete[] planarCodeHeader;
            continue;
        }

        // Create an adjacency matrix that will store THE COMPLEMENT of each graph in this file.
        bool** a = new bool*[nCurrent];
        for (unsigned i = 0; i < nCurrent; ++i) {
            a[i] = new bool[nCurrent];
        }

        // Now there can be any number of graphs encoded on the file, so we
        // give a unique ID to each one, starting at 1.
        unsigned isomerNumber = 1;

        // Each one of these potential graphs starts with a byte that
        // (interpreted as an unsigned integer) represents the number
        // of vertices (n). So while there is a byte to read in the file...
        char byte;
        while (inputFile.get(byte)) {

            // Start graph reading timer
            /*clock_t graphStart;
            clock_t graphEnd;
            cout << "Reading Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            // Reset values, remembering this will be THE COMPLEMENT
            for (unsigned i = 0; i < nCurrent; ++i) {
                for (unsigned j = 0; j < nCurrent; ++j) {
                    a[i][j] = true;
                }
            }

            unsigned numberOfVertices = (unsigned)((unsigned char)byte);

            // The next part of the file is divided into n sections. Each
            // section 1 <= v <= n contains the neighbours of vertex v,
            // presented in clockwise order, followed by a zero byte.
            for (unsigned v = 1; v <= numberOfVertices; ++v) {

                // Let w be the current neighbour of v being considered
                inputFile.get(byte);
                unsigned w = (unsigned)((unsigned char)byte);
                while (w != 0) {                // While there are neighbours to consider
                    a[v - 1][w - 1] = false;    // REMOVE the corresponding edge from the adjacency matrix (this will give us the complement)
                    inputFile.get(byte);        // Read the next neighbour
                    w = (unsigned)((unsigned char)byte);
                }
            }

            // Finished reading graph data.
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Start calculating
            /*cout << "Calculating Independence Number for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            // And pass it to the function that calculates the independence number
            // (actually it will calculate the size of a maximum clique of the complement)
            int alpha = independenceNumber(a, (int)nCurrent);

            // Calulation complete
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Update top/bottom 100 queues
            if (top100.size() < MAX_QUEUE_SIZE || alpha > top100.top().value) {
                if (top100.size() >= MAX_QUEUE_SIZE) {
                    top100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = alpha;
                top100.push(tpd);
            }
            if (bottom100.size() < MAX_QUEUE_SIZE || alpha < bottom100.top().value) {
                if (bottom100.size() >= MAX_QUEUE_SIZE) {
                    bottom100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = alpha;
                bottom100.push(tpd);
            }

            // Write results to file
            /*cout << "Writing results for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            outputFile << isomerNumber << "," << alpha << endl;

            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds."
                 << endl;*/

            // Increment the counter to be used as an ID for the next graph (if any).
            ++isomerNumber;
        }

        //
        // All the graphs in the file were processed, close it and destroy the auxiliary arrays.
        //
        inputFile.close();
        outputFile.close();
        delete[] planarCodeHeader;
        for (unsigned i = 0; i < nCurrent; ++i) {
            delete[] a[i];
        }
        delete[] a;

        //
        // Write to file the Top/Bottom 100 data corresponding to the graphs in this file
        //

        // Get how many numbers are actually on the queues and create vectors with them
        unsigned  countTop      = top100.size();
        unsigned  countBottom   = bottom100.size();
        vector<unsigned> topVector(countTop, 0);
        vector<unsigned> bottomVector(countBottom, 0);

        // Note that the pop method for the Top queues would return
        // numbers smallest > largest, and we want them largest > smallest
        for (unsigned i = countTop; i > 0; --i) {
            topVector[i - 1]    = top100.top().isomerID;
            top100.pop();
        }
        // For the Bottom queues, the opposite is true, but the solution is the same
        for (unsigned i = countBottom; i > 0; --i) {
            bottomVector[i - 1] = bottom100.top().isomerID;
            bottom100.pop();
        }

        // Now write the contents of the arrays to the correct columns on the output file...
        top100Bottom100File << nCurrent << ",\"";
        for (unsigned i = 0; i < countTop; ++i) {
            top100Bottom100File << topVector[i] << endl;
        }
        top100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countBottom; ++i) {
            top100Bottom100File << bottomVector[i] << endl;
        }
        top100Bottom100File << "\",\"";

        // Now sort the contents of the 4 arrays...
        sort(topVector.begin()   , topVector.end());
        sort(bottomVector.begin(), bottomVector.end());

        // And print the sorted data
        for (unsigned i = 0; i < countTop; ++i) {
            top100Bottom100File << topVector[i] << endl;
        }
        top100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countBottom; ++i) {
            top100Bottom100File << bottomVector[i] << endl;
        }
        top100Bottom100File << "\"" << endl;


        // Now reset the priority_queue objects to be used on the next file
        top100    = TopBottomQueue(TopBottomComparison(true)) ;
        bottom100 = TopBottomQueue() ;

        // Check time
        fileEnd = clock() ;
        cout << "Total time to process file: "
             << ( (double) ( fileEnd - fileStart ) ) / CLOCKS_PER_SEC 
             << " seconds." 
             << endl << endl;

    }

    // Close file
    top100Bottom100File.close();
  
    cout << "Finished.";
    if (errorDetected) {
        cout << " There were some errors, please review error log.";
    } else {
        cout << " Everything OK.";
    }
    cout << endl << endl;
  
    return EXIT_SUCCESS;

}
