/** 
 * \file testKekuleNumber.cpp
 *  
 * \brief A driver to test the implementation of the algorithm to calculate
 *        the number of kekule structures (perfect matchings) of a given fullerene
 *        graph.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date May 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
#include <iostream>                         // std::cout, std::endl
#include <fstream>                          // ifstream, ofstream
#include <sstream>                          // ostringstream
#include <string>                           // std::string
#include <cstring>                          // strcmp
#include <cstdlib>                          // exit, EXIT_SUCCESS, EXIT_FAILURE
#include <cstdio>                           // sprintf
#include <cmath>
#include <ctime>                            // time, clock, CLOCKS_PER_SEC, clock_t 
#include <iomanip>                          // std:setprecision
#include <limits>                           // std::numeric_limits
#include <ios>                              // std::fixed
//#include <cassert>                          // assert
#include <algorithm>                        // std:sort
#include <vector>                           // std::vector
#include <stack>
#include <queue>                            // std::priority_queue
#include <list>
#include <map>

#include "KekuleNumber.h"                    // 


// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

// Double output precision
typedef numeric_limits< double > dbl;

// Error flag
bool errorDetected = false;

// Struct to store each entry in Top 100 / Bottom 100 table and Comparison class
typedef struct TopBottomData {
    unsigned isomerID ;
    double   value    ;
} TopBottomData;

class TopBottomComparison {
    private:
        bool _reverse;

    public:
        TopBottomComparison(const bool& reverse = false) {
            _reverse = reverse;
        }

        bool operator() (const TopBottomData& lhs, const TopBottomData& rhs) const {
            if (_reverse) {
                return (lhs.value > rhs.value);
            } else {
                return (lhs.value < rhs.value);
            }
        }
};


/**
 * \fn inline bool isUnsigned(const string& s)
 *
 * \brief A function to test if a given string represents an unsigned integer number.
 *
 * \param s The string to be tested.
 *
 * \return True if the string represents an unsigned integer number, false otherwise.
 */
inline bool isUnsignedInteger(const string& s) {
    if (s.empty() || ((!isdigit(s[0])) && (s[0] != '+'))) {
        return false;
    }
    char* p;
    strtoul(s.c_str(), &p, 10);
    return (*p == 0);
}

/**
 * \fn void errorLog(const string& m)
 *
 * \brief A function to log an error message.
 *
 * \param m The message to be logged.
 */
void errorLog(const string& m) {
    // Set flag
    errorDetected = true;

    // Write to stdout
    cout << m << endl;

    // And to log file, with current time
    ofstream errorFile("../data/out/error/Error_Log.txt", ios_base::app);
    time_t now = time(0);
    struct tm  tstruct = *localtime(&now);
    char   buf[80];
    strftime(buf, sizeof(buf), "%d-%m-%Y.%X", &tstruct);
    errorFile << buf << ": " << m << endl;
    errorFile.close();
}

/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief A simple test of the implementation of the algorithm to calculate
 *        the number of kekule structures (perfect matchings) of a fullerene
 *        graph, using its dual as input.
 *
 * \param argc The number of command-line arguments.
 * \param argv The command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {

    //
    // Input validation
    //
    bool      isValidInput = false ;
    unsigned  nStart               ;
    unsigned  nFinish              ;

    if (argc == 3) {
        string arg1(argv[1]);
        string arg2(argv[2]);
        if (isUnsignedInteger(arg1) && isUnsignedInteger(arg2)) {
            nStart  = strtoul(argv[1], NULL, 10);
            nFinish = strtoul(argv[2], NULL, 10);
            if (nStart  >= 20     && nStart  != 22 && nStart  % 2 == 0 &&
                nFinish >= nStart && nFinish != 22 && nFinish % 2 == 0) {
                isValidInput = true;
            }
        }
    }

    if (!isValidInput) {
        cout << endl
             << "\t Usage: testKekuleNumber < arg1 >  < arg2 >" 
             << endl
             << "\t\t arg1   n at which to start calculating (an even integer >= 20 and != 22)"
             << endl
             << "\t\t arg2   n at which to stop calculating (an even integer >= arg1 and != 22)"
	         << endl
             << endl;

        return EXIT_FAILURE;
    }

    // Create a file to store the top 100 and bottom 100 isomers of each n...
    ofstream top100Bottom100File("../data/out/all-invariants/Kekule_Number.csv");
    if (!top100Bottom100File) {
        // Or log an error and exit the program.
        errorLog("Unable to create file: all-invariants/Kekule_Number.csv");
        return EXIT_FAILURE;
    }

    // And print the headers
    top100Bottom100File << "value_of_n,top_100_decreasing,bottom_100_increasing,top_100_by_id,bottom_100_by_id" << endl;

    // Now define 2 priority_queue objects (top and bottom 100).
    typedef priority_queue< TopBottomData, vector<TopBottomData>, TopBottomComparison > TopBottomQueue;
    TopBottomQueue top100(TopBottomComparison(true)); // The top 100 queues are inverse priority queues
    TopBottomQueue bottom100; // The bottom 100 queues are normal priority queues
    const unsigned MAX_QUEUE_SIZE = 100;

    //
    // For each value of N in the designated range...
    //
    for (unsigned nCurrent = nStart; nCurrent <= nFinish; nCurrent += 2) {

        // Skip if N is 22, as there are no fullerene graphs with 22 vertices
        if (nCurrent == 22) {
            continue;
        }

        ostringstream convert;
        convert << nCurrent;
        string nString = convert.str();
        if (nCurrent >= 132) {
            nString += "_ipr";
        }

        // File processing timers
        clock_t fileStart;
        clock_t fileEnd;
        cout << "Starting to process file Fullerenes_" << nCurrent << "..." << endl;
        fileStart = clock();

        // Open the corresponding input file
        string inputFileName = "../data/in/Fullerenes_" + nString;
        ifstream inputFile(inputFileName.c_str(), ifstream::binary);
        if (!inputFile) {
            // Or log an error and try to continue if unable to open
            errorLog("Unable to open file: " + inputFileName);
            continue;
        }

        // Create an output file (.csv format)...
        string outputFileName = "../data/out/kekule-number/Kekule_Number_" + nString + ".csv";
        ofstream outputFile(outputFileName.c_str());
        if (!outputFile) {
            // Or log an error and try to continue if unable to create.
            errorLog("Unable to create file: " + outputFileName);
            inputFile.close();
            continue;
        }

        // And print the headers of the data we will write.
        outputFile << "isomer_id,kekule_number" << endl;

        // Read the first 15 characters on the file and check if it
        // really is planar code
        char* planarCodeHeader = new char[15];
        inputFile.get(planarCodeHeader, 16); // one more than buffer size for 'get' to work
        if (strcmp(planarCodeHeader, ">>planar_code<<") != 0) {
            // Or log an error and try to continue if the header is wrong
            errorLog("File is missing planar code header: " + inputFileName + "    Header: " + planarCodeHeader);
            inputFile.close();
            outputFile.close();
            delete[] planarCodeHeader;
            continue;
        }

        // Create an OrderedVertexPair matrix that will be used (further below)
        // by all the graphs in the same file (they all have the same size, N)
        OrderedVertexPair** a = new OrderedVertexPair*[nCurrent + 1];
        for (unsigned i = 0; i <= nCurrent; ++i) {                                          // NOTE THAT THIS IS NOT ZERO BASED (the 0 is left empty)
            a[i] = new OrderedVertexPair[3];
        }

        // ... a Face map to store created faces
        map<unsigned, Face> faces;

        // ... an adjacency matrix that will be used to store the directions of each edge.
        // For the purposes of this experiment, the values of each position (i, j) will mean
        //  0 = no edge,
        //  2 = undirected edge
        //  1 = edge goes from i to j
        // -1 = edge goes from j to i
        const double NO_EDGE     = 0.0;
        const double UNDIRECTED  = 2.0;
        const double FROM_I_TO_J = 1.0;
        const double FROM_J_TO_I = -1.0;
        MatrixD pfaffianOrientation(nCurrent, nCurrent);                                    // AND THAT THIS IS ZERO BASED

        // ... an array of boolean flags to indicate which vertices were already
        // visited in the DFS that will take place when building tree T1
        bool* visited = new bool[nCurrent];                                                 // AND ALSO THIS

        // ... the actual stack used during that DFS
        stack<unsigned> dfs;

        // Now there can be any number of graphs encoded on the file, so we
        // give a unique ID to each one, starting at 1.
        unsigned isomerNumber = 1;

        // Each one of these potential graphs starts with a byte that
        // (interpreted as an unsigned integer) represents the number
        // of vertices (n). So while there is a byte to read in the file...
        char byte;
        while (inputFile.get(byte)) {

            // Start graph reading timer
            /*clock_t graphStart;
            clock_t graphEnd;
            cout << "Reading Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            unsigned numberOfVertices = (unsigned)((unsigned char)byte);

            // Remembering to reset everything to default values before use.
            for (unsigned i = 1; i <= nCurrent; ++i) {
                a[i][0].first  = i;
                a[i][0].second = 0;
                a[i][0].face   = 0;
                a[i][1].first  = i;
                a[i][1].second = 0;
                a[i][1].face   = 0;
                a[i][2].first  = i;
                a[i][2].second = 0;
                a[i][2].face   = 0;
            }
            for (unsigned i = 0; i < nCurrent; ++i) {
                for (unsigned j = 0; j < nCurrent; ++j) {
                    pfaffianOrientation(i, j) = NO_EDGE;
                }
                visited[i] = false;
            }

            // The next part of the file is divided into n sections. Each
            // section 1 <= v <= n contains the neighbours of vertex v,
            // presented in clockwise order, followed by a zero byte.
            for (unsigned v = 1; v <= numberOfVertices; ++v) {
                // Let w be the current neighbour of v being considered
                inputFile.get(byte);
                unsigned w = (unsigned)((unsigned char)byte);
                unsigned k = 0;
                while (w != 0) {                // While there are neighbours to consider
                    a[v][k].second = w;
                    pfaffianOrientation(v - 1, w - 1) = UNDIRECTED;
                    ++k;
                    inputFile.get(byte);        // Read the next neighbour
                    w = (unsigned)((unsigned char)byte);
                }
            }

            // Finished reading graph data.
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/


            // Start calculating
            /*cout << "Calculating Number of Kekule structures for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            // Find a spanning tree T1 of graph G using DFS
            // (same vertices as in G, edges in T1 are indicated by direction)
            dfs.push(0);
            visited[0] = true;
            while (!dfs.empty()) {
                unsigned v = dfs.top();
                dfs.pop();
                for (unsigned w = 0; w < nCurrent; ++w) {
                    // if there is an undirected edge between v and w, and w was not visited
                    if (pfaffianOrientation(v, w) == UNDIRECTED && !visited[w]) {
                        dfs.push(w);
                        visited[w] = true;
                        // Give a direction to this edge, indicating that it is now
                        // part of T1
                        pfaffianOrientation(v, w) = FROM_I_TO_J;
                        pfaffianOrientation(w, v) = FROM_J_TO_I;
                    }
                }
            }

            //
            // NOT SO SMALL DISCUSSION
            //
            // At this point, we should construct a tree (T2) having the same vertex set as the
            // dual of graph G, and having an edge between two vertices if the corresponding faces
            // in G share an edge NOT in T1 (i.e., share an undirected edge). Then we should remove
            // nodes from this tree one LEAF at a time, orienting the last unoriented edge on the
            // corresponding face so that there is an odd number of clockwise edges.
            // 
            // However, it seems the only point in actually building T2 is to know in which order
            // to consider the faces for orientation. If we didn't pick a face corresponding to a
            // leaf in T2, this face might have more than one unoriented edge, and that could "break"
            // the orientation on subsequente faces.
            //
            // So what we really need is just this order. And to find it, we don't need to build T2
            // (and then destroy it) at all. To do this, first we just pick an arbitrary face to be
            // the "root" of T2 (actually, its corresponding vertex in the dual), storing it in a list.
            // Then, we branch out from this face, "annexing" more faces (vertices in the dual). Instead
            // of actually inserting these in an actual "tree" object, we just append each one to the list.
            // After all faces are accounted for in this search, the list, when reversed, will give us
            // the order we want. That is, the last face inserted on the list corresponds surely to a leaf
            // in T2. After processing (and conceptually "removing" this face from T2), the face before that
            // will also correspond to a leaf, and so on and so forth, until we reach the first face (the
            // arbitrarily chosen root).
            //
            // END OF DISCUSSION
            //

            // So use the same mechanism as before (when building the dual)...
            unsigned faceID = 1;

            for (unsigned i = 1; i <= nCurrent; ++i) {
                for (unsigned j = 0; j < 3; ++j) {
                    if (a[i][j].face == 0) {
                        unsigned u = a[i][j].first;
                        unsigned v = a[i][j].second;
                        Face f = {faceID, false, &a[i][j]};
                        faces[faceID] = f;
                        a[i][j].face = faceID;
                        OrderedVertexPair* nextPair = nextF(a[i][j], a);
                        while (nextPair->first != u || nextPair->second != v) {
                            nextPair->face = faceID;
                            nextPair = nextF((*nextPair), a);
                        }
                        ++faceID;
                    }
                }
            }

            // Create the list that will be processed in reverse order further ahead.
            list<Face> t2Leaves;

            // Now, do a DFS starting at the vertex (in the dual) corresponding to
            // the first face (in the normal graph) and considering edges to exist
            // between vertices (in the dual) if, and only if, the corresponding faces
            // share an edge NOT int T1 (UNDIRECTED)

            dfs.push(1);
            //t2Leaves.push_back(faces[1]);     DON'T INSERT THIS FACE! IT WILL HAVE NO UNDIRECTED EDGES
            faces[1].visited = true;
            while (!dfs.empty()) {
                unsigned fID = dfs.top();
                dfs.pop();
                OrderedVertexPair* pair = faces[fID].pair;
                unsigned u = pair->first;
                unsigned v = pair->second;
                do {
                    if (pfaffianOrientation(pair->first - 1, pair->second - 1) == UNDIRECTED) {
                        OrderedVertexPair* other = symmetric((*pair), a);
                        unsigned neighbouringFaceID = other->face;
                        if (!faces[neighbouringFaceID].visited) {  //other->face is the neighbouring face on this edge
                            dfs.push(neighbouringFaceID);
                            t2Leaves.push_back(faces[neighbouringFaceID]);
                            faces[neighbouringFaceID].visited = true;
                        }
                    }
                    pair = nextF((*pair), a);
                } while (pair->first != u || pair->second != v);
                // End of neighbours of face F
            }


            // All we have to do now is iterate through t2Leaves backwards...
            //
            //  IGNORING THE ROOT OF T2 (i.e., the last face, which we DID NOT INSERT on t2Leaves)
            //
            for (list<Face>::reverse_iterator rit = t2Leaves.rbegin(); rit != t2Leaves.rend(); ++rit) {
                // Take each face ("leaf" on T2), and iterate through its bordering edges
                OrderedVertexPair* pair = (*rit).pair;
                unsigned u = pair->first;
                unsigned v = pair->second;
                // Note that nextF(u, v) gives back the edge (v, w) following (u, v) in counter-clockwise order
                // around their associated face. So to count the number of clockwise edges, we count how many
                // edges go from pair->second to pair->first.
                unsigned countClockwise  = 0;
                unsigned lastUndirectedU = 0;
                unsigned lastUndirectedV = 0;
                do {
                    if (pfaffianOrientation(pair->first - 1, pair->second - 1) == FROM_J_TO_I) {
                        //assert(pfaffianOrientation(pair->second - 1, pair->first - 1) == FROM_I_TO_J);
                        ++countClockwise;
                    } else if (pfaffianOrientation(pair->first - 1, pair->second - 1) == UNDIRECTED) {
                        //assert(lastUndirectedU == 0 && lastUndirectedV == 0);
                        lastUndirectedU = pair->first;
                        lastUndirectedV = pair->second;
                    }
                    pair = nextF((*pair), a);
                } while (pair->first != u || pair->second != v);

                // Now orient the last undirected edge around this face so that the number of
                // clockwise oriented edges is odd.
                // If the number is even...
                //assert(lastUndirectedU != 0 && lastUndirectedV != 0);
                if (countClockwise % 2 == 0) {
                    // Orient it clockwise
                    pfaffianOrientation(lastUndirectedU - 1, lastUndirectedV - 1) = FROM_J_TO_I;
                    pfaffianOrientation(lastUndirectedV - 1, lastUndirectedU - 1) = FROM_I_TO_J;
                } else {
                    // Orient it counterclockwise
                    pfaffianOrientation(lastUndirectedU - 1, lastUndirectedV - 1) = FROM_I_TO_J;
                    pfaffianOrientation(lastUndirectedV - 1, lastUndirectedU - 1) = FROM_J_TO_I;
                }
            }

            faces.clear();

            // Finally, calculate the number of perfect matchings
            // (the absolute value of the square root of the determinant of this oriented matrix)
            double kekule = abs(sqrt(pfaffianOrientation.determinant()));

            // Calulation complete
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Update top/bottom 100 queues
            if (top100.size() < MAX_QUEUE_SIZE || kekule > top100.top().value) {
                if (top100.size() >= MAX_QUEUE_SIZE) {
                    top100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = kekule;
                top100.push(tpd);
            }
            if (bottom100.size() < MAX_QUEUE_SIZE || kekule < bottom100.top().value) {
                if (bottom100.size() >= MAX_QUEUE_SIZE) {
                    bottom100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = kekule;
                bottom100.push(tpd);
            }

            // Write results to file
            /*cout << "Writing results for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            outputFile.precision(dbl::digits10);
            outputFile << isomerNumber << "," << scientific << kekule << endl;

            // Writing complete
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Increment the counter to be used as an ID for the next graph (if any).
            ++isomerNumber;
        }

        //
        // All the graphs in the file were processed, close it and destroy the auxiliary arrays.
        //
        inputFile.close();
        outputFile.close();
        delete[] planarCodeHeader;
        for (unsigned i = 0; i < nCurrent; ++i) {
            delete[] a[i];
        }
        delete[] a;
        delete[] visited;

        //
        // Write to file the Top/Bottom 100 data corresponding to the graphs in this file
        //

        // Get how many numbers are actually on the queues and create vectors with them
        unsigned  countTop      = top100.size();
        unsigned  countBottom   = bottom100.size();
        vector<unsigned> topVector(countTop, 0);
        vector<unsigned> bottomVector(countBottom, 0);

        // Note that the pop method for the Top queues would return
        // numbers smallest > largest, and we want them largest > smallest
        for (unsigned i = countTop; i > 0; --i) {
            topVector[i - 1]    = top100.top().isomerID;
            top100.pop();
        }
        // For the Bottom queues, the opposite is true, but the solution is the same
        for (unsigned i = countBottom; i > 0; --i) {
            bottomVector[i - 1] = bottom100.top().isomerID;
            bottom100.pop();
        }

        // Now write the contents of the arrays to the correct columns on the output file...
        top100Bottom100File << nCurrent << ",\"";
        for (unsigned i = 0; i < countTop; ++i) {
            top100Bottom100File << topVector[i] << endl;
        }
        top100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countBottom; ++i) {
            top100Bottom100File << bottomVector[i] << endl;
        }
        top100Bottom100File << "\",\"";

        // Now sort the contents of the 4 arrays...
        sort(topVector.begin()   , topVector.end());
        sort(bottomVector.begin(), bottomVector.end());

        // And print the sorted data
        for (unsigned i = 0; i < countTop; ++i) {
            top100Bottom100File << topVector[i] << endl;
        }
        top100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countBottom; ++i) {
            top100Bottom100File << bottomVector[i] << endl;
        }
        top100Bottom100File << "\"" << endl;


        // Now reset the priority_queue objects to be used on the next file
        top100    = TopBottomQueue(TopBottomComparison(true)) ;
        bottom100 = TopBottomQueue() ;

        // Check time
        fileEnd = clock() ;
        cout << "Total time to process file: "
             << ( (double) ( fileEnd - fileStart ) ) / CLOCKS_PER_SEC 
             << " seconds." 
             << endl << endl;

    }

    // Close file
    top100Bottom100File.close();
  
    cout << "Finished.";
    if (errorDetected) {
        cout << " There were some errors, please review error log.";
    } else {
        cout << " Everything OK.";
    }
    cout << endl << endl;
  
    return EXIT_SUCCESS;

}
