/** 
 * \file testPlanar2Dual.cpp
 *  
 * \brief A driver to test the implementation of the algorithm to calculate
 *        the dual of a given fullerene graph in planar code.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
#include <iostream>                         // std::cout, std::endl
#include <fstream>                          // ifstream, ofstream
#include <sstream>                          // ostringstream
#include <string>                           // std::string
#include <cstring>                          // strcmp
#include <cstdlib>                          // exit, EXIT_SUCCESS, EXIT_FAILURE
#include <cstdio>                           // sprintf
#include <ctime>                            // time, clock, CLOCKS_PER_SEC, clock_t 
#include <iomanip>                          // std:setprecision
#include <ios>                              // std::fixed
//#include <cassert>                          // assert
#include <algorithm>                        // std:sort
#include <vector>                           // std::vector

#include "Planar2Dual.h"                    // fowlerManolopoulosCriterion(hexNeighbourhoodIndices)


// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

// Error flag
bool errorDetected = false;


/**
 * \fn inline bool isUnsigned(const string& s)
 *
 * \brief A function to test if a given string represents an unsigned integer number.
 *
 * \param s The string to be tested.
 *
 * \return True if the string represents an unsigned integer number, false otherwise.
 */
inline bool isUnsignedInteger(const string& s) {
    if (s.empty() || ((!isdigit(s[0])) && (s[0] != '+'))) {
        return false;
    }
    char* p;
    strtoul(s.c_str(), &p, 10);
    return (*p == 0);
}

/**
 * \fn void errorLog(const string& m)
 *
 * \brief A function to log an error message.
 *
 * \param m The message to be logged.
 */
void errorLog(const string& m) {
    // Set flag
    errorDetected = true;

    // Write to stdout
    cout << m << endl;

    // And to log file, with current time
    ofstream errorFile("../data/out/error/Error_Log.txt", ios_base::app);
    time_t now = time(0);
    struct tm  tstruct = *localtime(&now);
    char   buf[80];
    strftime(buf, sizeof(buf), "%d-%m-%Y.%X", &tstruct);
    errorFile << buf << ": " << m << endl;
    errorFile.close();
}

/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief A simple test of the implementation of the algorithm to calculate
 *        the dual of a given fullerene graph.
 *
 * \param argc The number of command-line arguments.
 * \param argv The command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {

    //
    // Input validation
    //
    bool      isValidInput = false ;
    unsigned  nStart               ;
    unsigned  nFinish              ;

    if (argc == 3) {
        string arg1(argv[1]);
        string arg2(argv[2]);
        if (isUnsignedInteger(arg1) && isUnsignedInteger(arg2)) {
            nStart  = strtoul(argv[1], NULL, 10);
            nFinish = strtoul(argv[2], NULL, 10);
            if (nStart  >= 20     && nStart  != 22 && nStart  % 2 == 0 &&
                nFinish >= nStart && nFinish != 22 && nFinish % 2 == 0) {
                isValidInput = true;
            }
        }
    }

    if (!isValidInput) {
        cout << endl
             << "\t Usage: testPlanar2Dual < arg1 >  < arg2 >" 
             << endl
             << "\t\t arg1   n at which to start calculating (an even integer >= 20 and != 22)"
             << endl
             << "\t\t arg2   n at which to stop calculating (an even integer >= arg1 and != 22)"
	         << endl
             << endl;

        return EXIT_FAILURE;
    }

    //
    // For each value of N in the designated range...
    //
    for (unsigned nCurrent = nStart; nCurrent <= nFinish; nCurrent += 2) {

        // Skip if N is 22, as there are no fullerene graphs with 22 vertices
        if (nCurrent == 22) {
            continue;
        }

        ostringstream convert;
        convert << nCurrent;
        string nString = convert.str();
        if (nCurrent >= 132) {
            nString += "_ipr";
        }

        // File processing timers
        clock_t fileStart;
        clock_t fileEnd;
        cout << "Starting to process file Fullerenes_" << nCurrent << "..." << endl;
        fileStart = clock();

        // Open the corresponding input file
        string inputFileName = "../data/in/Fullerenes_" + nString;
        ifstream inputFile(inputFileName.c_str(), ifstream::binary);
        if (!inputFile) {
            // Or log an error and try to continue if unable to open
            errorLog("Unable to open file: " + inputFileName);
            continue;
        }

        // Create an output file (that will be used as input to another program)...
        string outputFileName = "../data/in/dual/Fullerenes_Dual_" + nString;
        ofstream outputFile(outputFileName.c_str());
        if (!outputFile) {
            // Or log an error and try to continue if unable to create.
            errorLog("Unable to create file: " + outputFileName);
            inputFile.close();
            continue;
        }

        // And print the planar code header
        outputFile << ">>planar_code<<";

        // Read the first 15 characters on the file and check if it
        // really is planar code
        char* planarCodeHeader = new char[15];
        inputFile.get(planarCodeHeader, 16); // one more than buffer size for 'get' to work
        if (strcmp(planarCodeHeader, ">>planar_code<<") != 0) {
            // Or log an error and try to continue if the header is wrong
            errorLog("File is missing planar code header: " + inputFileName + "    Header: " + planarCodeHeader);
            inputFile.close();
            outputFile.close();
            delete[] planarCodeHeader;
            continue;
        }

        // Create an OrderedVertexPair matrix that will be used (further below)
        // by all the graphs in the same file (they all have the same size, N)
        OrderedVertexPair** a = new OrderedVertexPair*[nCurrent + 1];
        for (unsigned i = 0; i <= nCurrent; ++i) {
            a[i] = new OrderedVertexPair[3];
        }

        // And a Face vector to store created faces
        vector<Face> faces;

        // Now there can be any number of graphs encoded on the file, so we
        // give a unique ID to each one, starting at 1.
        unsigned isomerNumber = 1;

        // Each one of these potential graphs starts with a byte that
        // (interpreted as an unsigned integer) represents the number
        // of vertices (n). So while there is a byte to read in the file...
        char byte;
        while (inputFile.get(byte)) {

            // Start graph reading timer
            /*clock_t graphStart;
            clock_t graphEnd;
            cout << "Reading Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            unsigned numberOfVertices = (unsigned)((unsigned char)byte);

            // Remembering to reset the matrix to default values before use.
            for (unsigned i = 1; i <= nCurrent; ++i) {
                a[i][0].first  = i;
                a[i][0].second = 0;
                a[i][0].face   = 0;
                a[i][1].first  = i;
                a[i][1].second = 0;
                a[i][1].face   = 0;
                a[i][2].first  = i;
                a[i][2].second = 0;
                a[i][2].face   = 0;
            }

            // The next part of the file is divided into n sections. Each
            // section 1 <= v <= n contains the neighbours of vertex v,
            // presented in clockwise order, followed by a zero byte.
            for (unsigned v = 1; v <= numberOfVertices; ++v) {
                // Let w be the current neighbour of v being considered
                inputFile.get(byte);
                unsigned w = (unsigned)((unsigned char)byte);
                unsigned k = 0;
                while (w != 0) {                // While there are neighbours to consider
                    a[v][k].second = w;
                    ++k;
                    inputFile.get(byte);        // Read the next neighbour
                    w = (unsigned)((unsigned char)byte);
                }
            }

            // Finished reading graph data.
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Start calculating
            /*cout << "Calculating Dual of Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            unsigned faceID = 1;

            for (unsigned i = 1; i <= nCurrent; ++i) {
                for (unsigned j = 0; j < 3; ++j) {
                    if (a[i][j].face == 0) {
                        unsigned u = a[i][j].first;
                        unsigned v = a[i][j].second;
                        Face f = {faceID, &a[i][j]};
                        faces.push_back(f);
                        a[i][j].face = faceID;
                        OrderedVertexPair* nextPair = nextF(a[i][j], a);
                        while (nextPair->first != u || nextPair->second != v) {
                            nextPair->face = faceID;
                            nextPair = nextF((*nextPair), a);
                        }
                        ++faceID;
                    }
                }
            }

            unsigned numberOfFaces = faces.size();
            unsigned char zeroByte = 0;
            outputFile << (unsigned char)numberOfFaces;

            for (unsigned i = 0; i < numberOfFaces; ++i) {
                Face f = faces[i];
                OrderedVertexPair* pair = f.pair;
                unsigned u = pair->first;
                unsigned v = pair->second;
                do {
                    OrderedVertexPair* other = symmetric((*pair), a);
                    outputFile << (unsigned char)other->face; // neighbour of f
                    pair = nextF((*pair), a);
                } while (pair->first != u || pair->second != v);
                outputFile << zeroByte;
                
            }

            faces.clear();

            // Calulation complete
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Increment the counter to be used as an ID for the next graph (if any).
            ++isomerNumber;
        }

        //
        // All the graphs in the file were processed, close it and destroy the auxiliary arrays.
        //
        inputFile.close();
        outputFile.close();
        delete[] planarCodeHeader;
        for (unsigned i = 0; i < nCurrent; ++i) {
            delete[] a[i];
        }
        delete[] a;

        // Check time
        fileEnd = clock() ;
        cout << "Total time to process file: "
             << ( (double) ( fileEnd - fileStart ) ) / CLOCKS_PER_SEC 
             << " seconds." 
             << endl << endl;

    }
  
    cout << "Finished.";
    if (errorDetected) {
        cout << " There were some errors, please review error log.";
    } else {
        cout << " Everything OK.";
    }
    cout << endl << endl;
  
    return EXIT_SUCCESS;

}
