/** 
 * \file testTaylorNumber.cpp
 *  
 * \brief A driver to test the implementation of the algorithm to calculate
 *        the taylor number of a given fullerene graph.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date May 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
#include <iostream>                         // std::cout, std::endl
#include <fstream>                          // ifstream, ofstream
#include <sstream>                          // ostringstream
#include <string>                           // std::string
#include <cstring>                          // strcmp
#include <cstdlib>                          // exit, EXIT_SUCCESS, EXIT_FAILURE
#include <cstdio>                           // sprintf
#include <cmath>
#include <ctime>                            // time, clock, CLOCKS_PER_SEC, clock_t 
#include <iomanip>                          // std:setprecision
#include <limits>                           // std::numeric_limits
#include <ios>                              // std::fixed
//#include <cassert>                          // assert
#include <algorithm>                        // std:sort
#include <vector>                           // std::vector
#include <stack>
#include <queue>                            // std::priority_queue
#include <list>
#include <map>

#include "TaylorNumber.h"                    // 


// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

// Double output precision
typedef numeric_limits< double > dbl;

// Error flag
bool errorDetected = false;

// Struct to store each entry in Top 100 / Bottom 100 table and Comparison class
typedef struct TopBottomData {
    unsigned isomerID ;
    double   value    ;
} TopBottomData;

class TopBottomComparison {
    private:
        bool _reverse;

    public:
        TopBottomComparison(const bool& reverse = false) {
            _reverse = reverse;
        }

        bool operator() (const TopBottomData& lhs, const TopBottomData& rhs) const {
            if (_reverse) {
                return (lhs.value > rhs.value);
            } else {
                return (lhs.value < rhs.value);
            }
        }
};


/**
 * \fn inline bool isUnsigned(const string& s)
 *
 * \brief A function to test if a given string represents an unsigned integer number.
 *
 * \param s The string to be tested.
 *
 * \return True if the string represents an unsigned integer number, false otherwise.
 */
inline bool isUnsignedInteger(const string& s) {
    if (s.empty() || ((!isdigit(s[0])) && (s[0] != '+'))) {
        return false;
    }
    char* p;
    strtoul(s.c_str(), &p, 10);
    return (*p == 0);
}

/**
 * \fn void errorLog(const string& m)
 *
 * \brief A function to log an error message.
 *
 * \param m The message to be logged.
 */
void errorLog(const string& m) {
    // Set flag
    errorDetected = true;

    // Write to stdout
    cout << m << endl;

    // And to log file, with current time
    ofstream errorFile("../data/out/error/Error_Log.txt", ios_base::app);
    time_t now = time(0);
    struct tm  tstruct = *localtime(&now);
    char   buf[80];
    strftime(buf, sizeof(buf), "%d-%m-%Y.%X", &tstruct);
    errorFile << buf << ": " << m << endl;
    errorFile.close();
}

/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief A simple test of the implementation of the algorithm to calculate
 *        the fries number of a fullerene graph.
 *
 * \param argc The number of command-line arguments.
 * \param argv The command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {

    //
    // Input validation
    //
    bool      isValidInput = false ;
    unsigned  nStart               ;
    unsigned  nFinish              ;
    string    specialFileName = "" ;

    if (argc == 3) {
        string arg1(argv[1]);
        string arg2(argv[2]);
        if (isUnsignedInteger(arg1) && isUnsignedInteger(arg2)) {
            nStart  = strtoul(argv[1], NULL, 10);
            nFinish = strtoul(argv[2], NULL, 10);
            if (nStart  >= 20     && nStart  != 22 && nStart  % 2 == 0 &&
                nFinish >= nStart && nFinish != 22 && nFinish % 2 == 0) {
                isValidInput = true;
            }
            specialFileName = arg1 + "_to_" + arg2;
        }
    }

    if (!isValidInput) {
        cout << endl
             << "\t Usage: testTaylorNumber < arg1 >  < arg2 >" 
             << endl
             << "\t\t arg1   n at which to start calculating (an even integer >= 20 and != 22)"
             << endl
             << "\t\t arg2   n at which to stop calculating (an even integer >= arg1 and != 22)"
	         << endl
             << endl;

        return EXIT_FAILURE;
    }

    // Create a file to store the top 100 and bottom 100 isomers of each n...
    ofstream top100Bottom100File("../data/out/all-invariants/Taylor_Number_" + specialFileName + ".csv");
    if (!top100Bottom100File) {
        // Or log an error and exit the program.
        errorLog("Unable to create file: all-invariants/Taylor_Number_" + specialFileName + ".csv");
        return EXIT_FAILURE;
    }

    // And print the headers
    top100Bottom100File << "value_of_n,top_100_decreasing,bottom_100_increasing,top_100_by_id,bottom_100_by_id" << endl;

    // Now define 2 priority_queue objects (top and bottom 100).
    typedef priority_queue< TopBottomData, vector<TopBottomData>, TopBottomComparison > TopBottomQueue;
    TopBottomQueue top100(TopBottomComparison(true)); // The top 100 queues are inverse priority queues
    TopBottomQueue bottom100; // The bottom 100 queues are normal priority queues
    const unsigned MAX_QUEUE_SIZE = 100;

    //
    // For each value of N in the designated range...
    //
    for (unsigned nCurrent = nStart; nCurrent <= nFinish; nCurrent += 2) {

        // Skip if N is 22, as there are no fullerene graphs with 22 vertices
        if (nCurrent == 22) {
            continue;
        }

        ostringstream convert;
        convert << nCurrent;
        string nString = convert.str();
        if (nCurrent >= 132) {
            nString += "_ipr";
        }

        // File processing timers
        clock_t fileStart;
        clock_t fileEnd;
        cout << "Starting to process file Fullerenes_" << nCurrent << "..." << endl;
        fileStart = clock();

        // Open the corresponding input file
        string inputFileName = "../data/in/Fullerenes_" + nString;
        ifstream inputFile(inputFileName.c_str(), ifstream::binary);
        if (!inputFile) {
            // Or log an error and try to continue if unable to open
            errorLog("Unable to open file: " + inputFileName);
            continue;
        }

        // Create an output file (.csv format)...
        string outputFileName = "../data/out/taylor-number/Taylor_Number_" + nString + ".csv";
        ofstream outputFile(outputFileName.c_str());
        if (!outputFile) {
            // Or log an error and try to continue if unable to create.
            errorLog("Unable to create file: " + outputFileName);
            inputFile.close();
            continue;
        }

        // And print the headers of the data we will write.
        outputFile << "isomer_id,taylor_number_a,tp_m_edges_in_pentagons" << endl;

        // Read the first 15 characters on the file and check if it
        // really is planar code
        char* planarCodeHeader = new char[15];
        inputFile.get(planarCodeHeader, 16); // one more than buffer size for 'get' to work
        if (strcmp(planarCodeHeader, ">>planar_code<<") != 0) {
            // Or log an error and try to continue if the header is wrong
            errorLog("File is missing planar code header: " + inputFileName + "    Header: " + planarCodeHeader);
            inputFile.close();
            outputFile.close();
            delete[] planarCodeHeader;
            continue;
        }

        // Create an OrderedVertexPair matrix that will be used (further below)
        // by all the graphs in the same file (they all have the same size, N)
        OrderedVertexPair** a = new OrderedVertexPair*[nCurrent + 1];
        for (unsigned i = 0; i <= nCurrent; ++i) {                                          // NOTE THAT THIS IS NOT ZERO BASED (the 0 is left empty)
            a[i] = new OrderedVertexPair[3];
        }

        // ... 3 arrays to store the non-zero coeffcients (for all graphs in the file)
        unsigned totalNonZeroCoefficients = ((27 * nCurrent) / 2) + 200;
        int* ia = new int[totalNonZeroCoefficients];
        int* ja = new int[totalNonZeroCoefficients];
        double* ar = new double[totalNonZeroCoefficients];
        unsigned totalNonZeroCoefficients2 = (21 * nCurrent) - 120 + 200;
        int* ia2 = new int[totalNonZeroCoefficients2];
        int* ja2 = new int[totalNonZeroCoefficients2];
        double* ar2 = new double[totalNonZeroCoefficients2];

        // ... a Face vector to store created faces
        //vector<Face> faces;

        // Now there can be any number of graphs encoded on the file, so we
        // give a unique ID to each one, starting at 1.
        unsigned isomerNumber = 1;

        // Each one of these potential graphs starts with a byte that
        // (interpreted as an unsigned integer) represents the number
        // of vertices (n). So while there is a byte to read in the file...
        char byte;
        while (inputFile.get(byte)) {

            // Start graph reading timer
            /*clock_t graphStart;
            clock_t graphEnd;
            cout << "Reading Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            unsigned numberOfVertices = (unsigned)((unsigned char)byte);

            // Remembering to reset everything to default values before use.
            for (unsigned i = 1; i <= nCurrent; ++i) {
                a[i][0].first          = i;
                a[i][0].second         = 0;
                a[i][0].face           = 0;
                a[i][0].edgeXvwVarID   = 0;
                a[i][0].edgePvwVarID   = 0;
                a[i][0].edgeAvwVarID   = 0;
                a[i][0].p2edgeXvwVarID = 0;
                a[i][0].p2edgePvwVarID = 0;
                a[i][0].p2edgeAvwVarID = 0;
                a[i][1].first          = i;
                a[i][1].second         = 0;
                a[i][1].face           = 0;
                a[i][1].edgeXvwVarID   = 0;
                a[i][1].edgePvwVarID   = 0;
                a[i][1].edgeAvwVarID   = 0;
                a[i][1].p2edgeXvwVarID = 0;
                a[i][1].p2edgePvwVarID = 0;
                a[i][1].p2edgeAvwVarID = 0;
                a[i][2].first          = i;
                a[i][2].second         = 0;
                a[i][2].face           = 0;
                a[i][2].edgeXvwVarID   = 0;
                a[i][2].edgePvwVarID   = 0;
                a[i][2].edgeAvwVarID   = 0;
                a[i][2].p2edgeXvwVarID = 0;
                a[i][2].p2edgePvwVarID = 0;
                a[i][2].p2edgeAvwVarID = 0;
            }
            for (unsigned i = 0; i < totalNonZeroCoefficients; ++i) {
                ia[i] = 0;
                ja[i] = 0;
                ar[i] = 0.0;
            }
            for (unsigned i = 0; i < totalNonZeroCoefficients2; ++i) {
                ia2[i] = 0;
                ja2[i] = 0;
                ar2[i] = 0.0;
            }

            //
            //
            // Create a new GLPK problem instance
            //
            //
            glp_prob *lp;
            lp = glp_create_prob();
            glp_term_out(GLP_OFF);
            glp_set_prob_name(lp, "Taylor Number Presolve");
            glp_set_obj_dir(lp, GLP_MIN);
            // Total auxiliary variables (restrictions) = 11n/2
            glp_add_rows(lp, (11 * nCurrent) / 2 );
            // Total structural variables = 9n/2
            glp_add_cols(lp, (9 * nCurrent) / 2 );
            
            unsigned variableID    = 1;
            unsigned restrictionID = 1;
            unsigned coefficientID = 0;
            unsigned variableID2    = 1;
            unsigned restrictionID2 = 1;
            unsigned coefficientID2 = 0;
            vector<unsigned> pentagonEdgesVariablesLp2;
            vector<ProblemRow> restrictionsLP2;
            vector<ProblemColumn> variablesLP2;

            // The next part of the file is divided into n sections. Each
            // section 1 <= v <= n contains the neighbours of vertex v,
            // presented in clockwise order, followed by a zero byte.
            for (unsigned v = 1; v <= numberOfVertices; ++v) {
                // Let w be the current neighbour of v being considered
                inputFile.get(byte);
                unsigned w = (unsigned)((unsigned char)byte);
                unsigned k = 0;
                while (w != 0) {                // While there are neighbours to consider
                    a[v][k].second = w;
                    if (v < w) {
                        // Create a new structural binary variable
                        // GLPK
                        glp_set_col_name(lp, variableID, "Xvw");
                        glp_set_col_bnds(lp, variableID, GLP_DB, 0.0, 1.0);
                        glp_set_obj_coef(lp, variableID, 0.0);
                        glp_set_col_kind(lp, variableID, GLP_BV);
                        a[v][k].edgeXvwVarID = variableID;
                        ++variableID;

                        // Create two new binary variables, Pvw, Avw
                        // Assume this edge won't be part of a pentagon, so Pvw starts at 0 (if it is, it's changed later)
                        glp_set_col_name(lp, variableID, "Pvw");
                        glp_set_col_bnds(lp, variableID, GLP_FX, 0.0, 0.0);
                        glp_set_obj_coef(lp, variableID, 0.0);
                        a[v][k].edgePvwVarID = variableID;
                        ++variableID;
                        glp_set_col_name(lp, variableID, "Avw");
                        glp_set_col_bnds(lp, variableID, GLP_DB, 0.0, 1.0);
                        glp_set_obj_coef(lp, variableID, 1.0);
                        glp_set_col_kind(lp, variableID, GLP_BV);
                        a[v][k].edgeAvwVarID = variableID;
                        ++variableID;

                        // Create 3 new restrictions on Avw, Pvw, and Xvw
                        // Aij <= Xij   ->   Aij - Xij <= 0  ->  Aij - Xij = Q <= 0
                        glp_set_row_name(lp, restrictionID, "Q");
                        glp_set_row_bnds(lp, restrictionID, GLP_UP, 0.0, 0.0);
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgeAvwVarID;
                        ar[coefficientID] = 1.0;
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgeXvwVarID;
                        ar[coefficientID] = -1.0;
                        ++restrictionID;
                        // Aij <= Pij   ->   Aij - Pij <= 0  ->  Aij - Pij = Q <= 0
                        glp_set_row_name(lp, restrictionID, "Q");
                        glp_set_row_bnds(lp, restrictionID, GLP_UP, 0.0, 0.0);
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgeAvwVarID;
                        ar[coefficientID] = 1.0;
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgePvwVarID;
                        ar[coefficientID] = -1.0;
                        ++restrictionID;
                        // Aij + 1 >= Xij + Pij   ->    Aij - Xij - Pij = Q >= -1
                        glp_set_row_name(lp, restrictionID, "Q");
                        glp_set_row_bnds(lp, restrictionID, GLP_LO, -1.0, 0.0);
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgeAvwVarID;
                        ar[coefficientID] = 1.0;
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgeXvwVarID;
                        ar[coefficientID] = -1.0;
                        ++coefficientID;
                        ia[coefficientID] = restrictionID;
                        ja[coefficientID] = a[v][k].edgePvwVarID;
                        ar[coefficientID] = -1.0;
                        ++restrictionID;

                        //
                        // Replicate for lp2
                        //
                        ProblemColumn columnXvw;            // Xvw
                        columnXvw.id = variableID2;
                        columnXvw.boundType = GLP_DB;
                        columnXvw.lowerBound = 0.0;
                        columnXvw.upperBound = 1.0;
                        columnXvw.coefficient = 0.0;
                        columnXvw.kind = GLP_BV;
                        variablesLP2.push_back(columnXvw);
                        a[v][k].p2edgeXvwVarID = variableID2;
                        ++variableID2;
                        ProblemColumn columnPvw;            // Pvw
                        columnPvw.id = variableID2;
                        columnPvw.boundType = GLP_FX;
                        columnPvw.lowerBound = 0.0;
                        columnPvw.upperBound = 0.0;
                        columnPvw.coefficient = 0.0;
                        columnPvw.kind = GLP_IV;
                        variablesLP2.push_back(columnPvw);
                        a[v][k].p2edgePvwVarID = variableID2;
                        ++variableID2;
                        ProblemColumn columnAvw;            // Avw
                        columnAvw.id = variableID2;
                        columnAvw.boundType = GLP_DB;
                        columnAvw.lowerBound = 0.0;
                        columnAvw.upperBound = 1.0;
                        columnAvw.coefficient = 0.0;
                        columnAvw.kind = GLP_BV;
                        variablesLP2.push_back(columnAvw);
                        a[v][k].p2edgeAvwVarID = variableID2;
                        ++variableID2;
                        pentagonEdgesVariablesLp2.push_back(a[v][k].p2edgeAvwVarID);
                        ProblemRow rowQ1;
                        rowQ1.id = restrictionID2;
                        rowQ1.boundType = GLP_UP;
                        rowQ1.lowerBound = 0.0;
                        rowQ1.upperBound = 0.0;
                        restrictionsLP2.push_back(rowQ1);
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgeAvwVarID;
                        ar2[coefficientID2] = 1.0;
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgeXvwVarID;
                        ar2[coefficientID2] = -1.0;
                        ++restrictionID2;
                        ProblemRow rowQ2;
                        rowQ2.id = restrictionID2;
                        rowQ2.boundType = GLP_UP;
                        rowQ2.lowerBound = 0.0;
                        rowQ2.upperBound = 0.0;
                        restrictionsLP2.push_back(rowQ2);
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgeAvwVarID;
                        ar2[coefficientID2] = 1.0;
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgePvwVarID;
                        ar2[coefficientID2] = -1.0;
                        ++restrictionID2;
                        ProblemRow rowQ3;
                        rowQ3.id = restrictionID2;
                        rowQ3.boundType = GLP_LO;
                        rowQ3.lowerBound = -1.0;
                        rowQ3.upperBound = 0.0;
                        restrictionsLP2.push_back(rowQ3);
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgeAvwVarID;
                        ar2[coefficientID2] = 1.0;
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgeXvwVarID;
                        ar2[coefficientID2] = -1.0;
                        ++coefficientID2;
                        ia2[coefficientID2] = restrictionID2;
                        ja2[coefficientID2] = a[v][k].p2edgePvwVarID;
                        ar2[coefficientID2] = -1.0;
                        ++restrictionID2;

                    } else {
                        // GLPK
                        if (a[w][0].second == v) {
                            a[v][k].edgeXvwVarID = a[w][0].edgeXvwVarID;
                            a[v][k].edgePvwVarID = a[w][0].edgePvwVarID;
                            a[v][k].edgeAvwVarID = a[w][0].edgeAvwVarID;

                            a[v][k].p2edgeXvwVarID = a[w][0].p2edgeXvwVarID;
                            a[v][k].p2edgePvwVarID = a[w][0].p2edgePvwVarID;
                            a[v][k].p2edgeAvwVarID = a[w][0].p2edgeAvwVarID;
                        } else if (a[w][1].second == v) {
                            a[v][k].edgeXvwVarID = a[w][1].edgeXvwVarID;
                            a[v][k].edgePvwVarID = a[w][1].edgePvwVarID;
                            a[v][k].edgeAvwVarID = a[w][1].edgeAvwVarID;

                            a[v][k].p2edgeXvwVarID = a[w][1].p2edgeXvwVarID;
                            a[v][k].p2edgePvwVarID = a[w][1].p2edgePvwVarID;
                            a[v][k].p2edgeAvwVarID = a[w][1].p2edgeAvwVarID;
                        } else {
                            a[v][k].edgeXvwVarID = a[w][2].edgeXvwVarID;
                            a[v][k].edgePvwVarID = a[w][2].edgePvwVarID;
                            a[v][k].edgeAvwVarID = a[w][2].edgeAvwVarID;

                            a[v][k].p2edgeXvwVarID  = a[w][2].p2edgeXvwVarID;
                            a[v][k].p2edgePvwVarID = a[w][2].p2edgePvwVarID;
                            a[v][k].p2edgeAvwVarID = a[w][2].p2edgeAvwVarID;
                        }
                    }
                    ++k;
                    inputFile.get(byte);        // Read the next neighbour
                    w = (unsigned)((unsigned char)byte);
                }
            }

            // Finished reading graph data.
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Start calculating
            /*cout << "Calculating Taylor Number for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            // So use the same mechanism as before (when building the dual)...
            unsigned faceID        = 1;

            for (unsigned i = 1; i <= nCurrent; ++i) {
                for (unsigned j = 0; j < 3; ++j) {
                    if (a[i][j].face == 0) {
                        unsigned u = a[i][j].first;
                        unsigned v = a[i][j].second;
                        //Face f = {faceID, &a[i][j]};
                        //faces.push_back(f);
                        a[i][j].face = faceID;
                        OrderedVertexPair* nextPair = nextF(a[i][j], a);
                        unsigned sides = 1;
                        while (nextPair->first != u || nextPair->second != v) {
                            nextPair->face = faceID;
                            nextPair = nextF((*nextPair), a);
                            ++sides;
                        }

                        // If this face is a hexagon
                        if (sides == 6) {
                            //
                            // For lp2 only
                            //
                            ProblemColumn columnYs;
                            columnYs.id = variableID2;
                            columnYs.boundType = GLP_DB;
                            columnYs.lowerBound = 0.0;
                            columnYs.upperBound = 1.0;
                            columnYs.coefficient = 1.0;
                            columnYs.kind = GLP_BV;
                            variablesLP2.push_back(columnYs);
                            unsigned ys = variableID2;
                            ++variableID2;
                            ProblemColumn columnZs;
                            columnZs.id = variableID2;
                            columnZs.boundType = GLP_DB;
                            columnZs.lowerBound = 0.0;
                            columnZs.upperBound = 1.0;
                            columnZs.coefficient = 1.0;
                            columnZs.kind = GLP_BV;
                            variablesLP2.push_back(columnZs);
                            unsigned zs = variableID2;
                            ++variableID2;

                            ProblemRow rowQ1;
                            rowQ1.id = restrictionID2;
                            rowQ1.boundType = GLP_DB;
                            rowQ1.lowerBound = 0.0;
                            rowQ1.upperBound = 1.0;
                            restrictionsLP2.push_back(rowQ1);
                            ++coefficientID2;
                            ia2[coefficientID2] = restrictionID2;
                            ja2[coefficientID2] = ys;
                            ar2[coefficientID2] = -1.0;
                            ++coefficientID2;
                            ia2[coefficientID2] = restrictionID2;
                            ja2[coefficientID2] = a[i][j].p2edgeXvwVarID;
                            ar2[coefficientID2] = 1.0;
                            ++restrictionID2;
                            bool alternate = true;
                            nextPair = nextF(a[i][j], a);
                            while (nextPair->first != u || nextPair->second != v) {
                                ProblemRow rowQothers;
                                rowQothers.id = restrictionID2;
                                rowQothers.boundType = GLP_DB;
                                rowQothers.lowerBound = 0.0;
                                rowQothers.upperBound = 1.0;
                                restrictionsLP2.push_back(rowQothers);
                                ++coefficientID2;
                                ia2[coefficientID2] = restrictionID2;
                                if (alternate) {
                                    ja2[coefficientID2] = zs;
                                    alternate = false;
                                } else {
                                    ja2[coefficientID2] = ys;
                                    alternate = true;
                                }
                                ar2[coefficientID2] = -1.0;
                                ++coefficientID2;
                                ia2[coefficientID2] = restrictionID2;
                                ja2[coefficientID2] = nextPair->p2edgeXvwVarID;
                                ar2[coefficientID2] = 1.0;
                                ++restrictionID2;
                                nextPair = nextF((*nextPair), a);
                            }
                        } else {
                            // Face is a pentagon, change Pvw of all the edges to 1
                            glp_set_col_bnds(lp, a[i][j].edgePvwVarID, GLP_FX, 1.0, 1.0);
                            unsigned currentVariablesLP2Size = variablesLP2.size();
                            for (unsigned it = 0; it < currentVariablesLP2Size; ++it) {
                                if (variablesLP2[it].id == a[i][j].p2edgePvwVarID) {
                                    variablesLP2[it].lowerBound = 1.0;
                                    variablesLP2[it].upperBound = 1.0;
                                }
                            }
                            nextPair = nextF(a[i][j], a);
                            while (nextPair->first != u || nextPair->second != v) {
                                glp_set_col_bnds(lp, nextPair->edgePvwVarID, GLP_FX, 1.0, 1.0);
                                for (unsigned it = 0; it < currentVariablesLP2Size; ++it) {
                                    if (variablesLP2[it].id == nextPair->p2edgePvwVarID) {
                                        variablesLP2[it].lowerBound = 1.0;
                                        variablesLP2[it].upperBound = 1.0;
                                    }
                                }
                                nextPair = nextF((*nextPair), a);
                            }
                        }
                        ++faceID;
                    }
                }

                // For each vertex, create a restriction P = Xij + Xik + Xil (variables associated with edges incident on i)
                glp_set_row_name(lp, restrictionID, "p");
                glp_set_row_bnds(lp, restrictionID, GLP_FX, 1.0, 1.0);
                // Create coefficients: i is the line (restrictionID), j is the column (variableID), r is the coefficient value
                ++coefficientID;
                ia[coefficientID] = restrictionID;
                ja[coefficientID] = a[i][0].edgeXvwVarID;
                ar[coefficientID] = 1.0;
                ++coefficientID;
                ia[coefficientID] = restrictionID;
                ja[coefficientID] = a[i][1].edgeXvwVarID;
                ar[coefficientID] = 1.0;
                ++coefficientID;
                ia[coefficientID] = restrictionID;
                ja[coefficientID] = a[i][2].edgeXvwVarID;
                ar[coefficientID] = 1.0;
                // Increment restriction afterwards
                ++restrictionID;

                //
                // Replicate for lp2
                //
                ProblemRow rowP;
                rowP.id = restrictionID2;
                rowP.boundType = GLP_FX;
                rowP.lowerBound = 1.0;
                rowP.upperBound = 1.0;
                restrictionsLP2.push_back(rowP);
                ++coefficientID2;
                ia2[coefficientID2] = restrictionID2;
                ja2[coefficientID2] = a[i][0].p2edgeXvwVarID;
                ar2[coefficientID2] = 1.0;
                ++coefficientID2;
                ia2[coefficientID2] = restrictionID2;
                ja2[coefficientID2] = a[i][1].p2edgeXvwVarID;
                ar2[coefficientID2] = 1.0;
                ++coefficientID2;
                ia2[coefficientID2] = restrictionID2;
                ja2[coefficientID2] = a[i][2].p2edgeXvwVarID;
                ar2[coefficientID2] = 1.0;
                ++restrictionID2;
            }

            //faces.clear();

            //
            // Fine tune parameters
            //            
            glp_smcp sparam;
            glp_init_smcp(&sparam);
            sparam.meth = GLP_DUALP; // default = GLP_PRIMAL, meh plus = GLP_DUAL, GLP_DUALP
            sparam.pricing = GLP_PT_STD; // default = GLP_PT_PSE, meh plus = GLP_PT_STD
            sparam.r_test = GLP_RT_STD; // default = GLP_RT_HAR, meh plus = GLP_RT_STD

            glp_iocp param;
            glp_init_iocp(&param);
            param.br_tech = GLP_BR_FFV; // bestest = GLP_BR_FFV, best = GLP_BR_MFV, good = GLP_BR_LFV, default = GLP_BR_DTH, worse = GLP_BR_PCH, 
            param.bt_tech = GLP_BT_BLB; // default = GLP_BT_BLB, meh= GLP_BT_DFS, GLP_BT_BPH, GLP_BT_BFS
            param.pp_tech = GLP_PP_NONE; // slightly = GLP_PP_NONE , even less slightly = GLP_PP_ROOT,  default = GLP_PP_ALL

            //
            // GLPK Calculate the "pre" taylor number, Tp (number of matched edges on pentagons)
            //
            glp_load_matrix(lp, coefficientID, ia, ja, ar);
            glp_simplex(lp, &sparam);
            glp_intopt(lp, &param);
            double numberOfMatchingEdgesOnPentagons = glp_mip_obj_val(lp);

            //
            // GLPK cleanup
            //
            glp_delete_prob(lp);
            glp_free_env();

            //
            //
            // Create a new GLPK problem instance
            //
            //
            glp_prob *lp2;

            lp2 = glp_create_prob();
            glp_term_out(GLP_OFF);
            glp_set_prob_name(lp2, "Taylor Number");
            glp_set_obj_dir(lp2, GLP_MAX);
            // Total auxiliary variables (restrictions) = 17n/2 - 60 + 1
            glp_add_rows(lp2,     ((17 * nCurrent) / 2)  - 60 + 1    );
            // Total structural variables = 11n/2 - 20
            glp_add_cols(lp2,     ((11 * nCurrent) / 2)  - 20   );

            for (vector<ProblemColumn>::iterator it = variablesLP2.begin(); it != variablesLP2.end(); ++it) {
                ProblemColumn c = (*it);
                glp_set_col_name(lp2, c.id, "Var");
                glp_set_col_bnds(lp2, c.id, c.boundType, c.lowerBound, c.upperBound);
                glp_set_obj_coef(lp2, c.id, c.coefficient);
                glp_set_col_kind(lp2, c.id, c.kind);
            }

            for (vector<ProblemRow>::iterator it = restrictionsLP2.begin(); it != restrictionsLP2.end(); ++it) {
                ProblemRow r = (*it);
                glp_set_row_name(lp2, r.id, "Rest");
                glp_set_row_bnds(lp2, r.id, r.boundType, r.lowerBound, r.upperBound);
            }

            // And use the previously calculated Tp to create the final restriction on the actual taylor number calculation
            glp_set_row_name(lp2, restrictionID2, "Tp");
            glp_set_row_bnds(lp2, restrictionID2, GLP_FX, numberOfMatchingEdgesOnPentagons, numberOfMatchingEdgesOnPentagons);
            for (vector<unsigned>::iterator it = pentagonEdgesVariablesLp2.begin(); it != pentagonEdgesVariablesLp2.end(); ++it) {
                ++coefficientID2;
                ia2[coefficientID2] = restrictionID2;
                ja2[coefficientID2] = (*it);
                ar2[coefficientID2] = 1.0;
            }
            ++restrictionID2;

            // Fine tune parameters
            glp_smcp sparam2;
            glp_init_smcp(&sparam2);
            sparam2.meth = GLP_PRIMAL; // default = GLP_PRIMAL, meh = GLP_DUAL, GLP_DUALP
            sparam2.pricing = GLP_PT_PSE; // default = GLP_PT_PSE, worse = GLP_PT_STD
            sparam2.r_test = GLP_RT_HAR; // default = GLP_RT_HAR, meh = GLP_RT_STD

            glp_iocp param2;
            glp_init_iocp(&param2);
            param2.br_tech = GLP_BR_MFV; // bestest = GLP_BR_MFV , best = GLP_BR_FFV, good = GLP_BR_LFV, default = GLP_BR_DTH, worse = GLP_BR_PCH, 
            param2.bt_tech = GLP_BT_BLB; // default = GLP_BT_BLB, meh= GLP_BT_DFS, GLP_BT_BPH, GLP_BT_BFS
            param2.pp_tech = GLP_PP_NONE; // slightly = GLP_PP_NONE , even less slightly = GLP_PP_ROOT,  default = GLP_PP_ALL

            glp_load_matrix(lp2, coefficientID2, ia2, ja2, ar2);
            glp_simplex(lp2, &sparam2);
            glp_intopt(lp2, &param2);
            double taylor = glp_mip_obj_val(lp2);

            //
            // GLPK cleanup
            //
            glp_delete_prob(lp2);
            glp_free_env();

            // Update top/bottom 100 queues
            if (top100.size() < MAX_QUEUE_SIZE || taylor > top100.top().value) {
                if (top100.size() >= MAX_QUEUE_SIZE) {
                    top100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = taylor;
                top100.push(tpd);
            }
            if (bottom100.size() < MAX_QUEUE_SIZE || taylor < bottom100.top().value) {
                if (bottom100.size() >= MAX_QUEUE_SIZE) {
                    bottom100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = taylor;
                bottom100.push(tpd);
            }

            // Write results to file
            /*cout << "Writing results for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            outputFile.precision(dbl::digits10);
            outputFile << isomerNumber << "," << scientific << taylor << "," << scientific << numberOfMatchingEdgesOnPentagons << endl;

            // Writing complete
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Increment the counter to be used as an ID for the next graph (if any).
            ++isomerNumber;
        }

        //
        // All the graphs in the file were processed, close it and destroy the auxiliary arrays.
        //
        inputFile.close();
        outputFile.close();
        delete[] planarCodeHeader;
        for (unsigned i = 0; i < nCurrent; ++i) {
            delete[] a[i];
        }
        delete[] a;
        delete[] ia;
        delete[] ja;
        delete[] ar;
        delete[] ia2;
        delete[] ja2;
        delete[] ar2;

        //
        // Write to file the Top/Bottom 100 data corresponding to the graphs in this file
        //

        // Get how many numbers are actually on the queues and create vectors with them
        unsigned  countTop      = top100.size();
        unsigned  countBottom   = bottom100.size();
        vector<unsigned> topVector(countTop, 0);
        vector<unsigned> bottomVector(countBottom, 0);

        // Note that the pop method for the Top queues would return
        // numbers smallest > largest, and we want them largest > smallest
        for (unsigned i = countTop; i > 0; --i) {
            topVector[i - 1]    = top100.top().isomerID;
            top100.pop();
        }
        // For the Bottom queues, the opposite is true, but the solution is the same
        for (unsigned i = countBottom; i > 0; --i) {
            bottomVector[i - 1] = bottom100.top().isomerID;
            bottom100.pop();
        }

        // Now write the contents of the arrays to the correct columns on the output file...
        top100Bottom100File << nCurrent << ",\"";
        for (unsigned i = 0; i < countTop; ++i) {
            top100Bottom100File << topVector[i] << endl;
        }
        top100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countBottom; ++i) {
            top100Bottom100File << bottomVector[i] << endl;
        }
        top100Bottom100File << "\",\"";

        // Now sort the contents of the 4 arrays...
        sort(topVector.begin()   , topVector.end());
        sort(bottomVector.begin(), bottomVector.end());

        // And print the sorted data
        for (unsigned i = 0; i < countTop; ++i) {
            top100Bottom100File << topVector[i] << endl;
        }
        top100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countBottom; ++i) {
            top100Bottom100File << bottomVector[i] << endl;
        }
        top100Bottom100File << "\"" << endl;


        // Now reset the priority_queue objects to be used on the next file
        top100    = TopBottomQueue(TopBottomComparison(true)) ;
        bottom100 = TopBottomQueue() ;

        // Check time
        fileEnd = clock() ;
        cout << "Total time to process file: "
             << ( (double) ( fileEnd - fileStart ) ) / CLOCKS_PER_SEC 
             << " seconds." 
             << endl << endl;

    }

    // Close file
    top100Bottom100File.close();
  
    cout << "Finished.";
    if (errorDetected) {
        cout << " There were some errors, please review error log.";
    } else {
        cout << " Everything OK.";
    }
    cout << endl << endl;
  
    return EXIT_SUCCESS;

}
