/** 
 * \file testWienerDiameter.cpp
 *  
 * \brief A driver to test the implementation of the algorithm to calculate
 *        the Wiener Index and the Diameter of a fullerene graph.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
#include <iostream>                         // std::cout, std::endl
#include <fstream>                          // ifstream, ofstream
#include <sstream>                          // ostringstream
#include <string>                           // std::string
#include <cstring>                          // strcmp
#include <cstdlib>                          // exit, EXIT_SUCCESS, EXIT_FAILURE
#include <cstdio>                           // sprintf
#include <ctime>                            // time, clock, CLOCKS_PER_SEC, clock_t 
#include <iomanip>                          // std:setprecision
#include <ios>                              // std::fixed
//#include <cassert>                          // assert
#include <algorithm>                        // std:sort
#include <vector>                           // std::vector
#include <queue>                            // std::priority_queue

#include "WienerDiameter.h"                 // wienerIndexAndDiameter(distances, n, diameter)


// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

// Error flag
bool errorDetected = false;

// Struct to store each entry in Top 100 / Bottom 100 table and Comparison class
typedef struct TopBottomData {
    unsigned isomerID ;
    unsigned value    ;
} TopBottomData;

class TopBottomComparison {
    private:
        bool _reverse;

    public:
        TopBottomComparison(const bool& reverse = false) {
            _reverse = reverse;
        }

        bool operator() (const TopBottomData& lhs, const TopBottomData& rhs) const {
            if (_reverse) {
                return (lhs.value > rhs.value);
            } else {
                return (lhs.value < rhs.value);
            }
        }
};

/**
 * \fn inline bool isUnsigned(const string& s)
 *
 * \brief A function to test if a given string represents an unsigned integer number.
 *
 * \param s The string to be tested.
 *
 * \return True if the string represents an unsigned integer number, false otherwise.
 */
inline bool isUnsignedInteger(const string& s) {
    if (s.empty() || ((!isdigit(s[0])) && (s[0] != '+'))) {
        return false;
    }
    char* p;
    strtoul(s.c_str(), &p, 10);
    return (*p == 0);
}

/**
 * \fn void errorLog(const string& m)
 *
 * \brief A function to log an error message.
 *
 * \param m The message to be logged.
 */
void errorLog(const string& m) {
    // Set flag
    errorDetected = true;

    // Write to stdout
    cout << m << endl;

    // And to log file, with current time
    ofstream errorFile("../data/out/error/Error_Log.txt", ios_base::app);
    time_t now = time(0);
    struct tm  tstruct = *localtime(&now);
    char   buf[80];
    strftime(buf, sizeof(buf), "%d-%m-%Y.%X", &tstruct);
    errorFile << buf << ": " << m << endl;
    errorFile.close();
}

/**
 * \fn int main( int argc , char* argv[] )
 *
 * \brief A simple test of the implementation of the algorithm to calculate
 *        the Wiener Index and the Diameter of a fullerene graph.
 *
 * \param argc The number of command-line arguments.
 * \param argv The command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {

    //
    // Input validation
    //
    bool      isValidInput = false ;
    unsigned  nStart               ;
    unsigned  nFinish              ;

    if (argc == 3) {
        string arg1(argv[1]);
        string arg2(argv[2]);
        if (isUnsignedInteger(arg1) && isUnsignedInteger(arg2)) {
            nStart  = strtoul(argv[1], NULL, 10);
            nFinish = strtoul(argv[2], NULL, 10);
            if (nStart  >= 20     && nStart  != 22 && nStart  % 2 == 0 &&
                nFinish >= nStart && nFinish != 22 && nFinish % 2 == 0) {
                isValidInput = true;
            }
        }
    }

    if (!isValidInput) {
        cout << endl
             << "\t Usage: testWienerDiameter < arg1 >  < arg2 >" 
             << endl
             << "\t\t arg1   n at which to start calculating (an even integer >= 20 and != 22)"
             << endl
             << "\t\t arg2   n at which to stop calculating (an even integer >= arg1 and != 22)"
	         << endl
             << endl;

        return EXIT_FAILURE;
    }

    // Create two files to store the top 100 and bottom 100 isomers of each n (both for the wiener index and for diameters)...
    ofstream wienerIndexTop100Bottom100File("../data/out/all-invariants/Wiener_Index.csv");
    ofstream diameterTop100Bottom100File("../data/out/all-invariants/Diameter.csv");
    if (!wienerIndexTop100Bottom100File || !diameterTop100Bottom100File) {
        // Or log an error and exit the program.
        errorLog("Unable to create files: all-invariants/Wiener_Index.csv   all-invariants/Diameter.csv");
        return EXIT_FAILURE;
    }

    // And print the headers
    wienerIndexTop100Bottom100File << "value_of_n,top_100_decreasing,bottom_100_increasing,top_100_by_id,bottom_100_by_id" << endl;
    diameterTop100Bottom100File    << "value_of_n,top_100_decreasing,bottom_100_increasing,top_100_by_id,bottom_100_by_id" << endl;

    // And define 4 priority_queue objects (top and bottom 100 for wiener index and diameter).
    typedef priority_queue< TopBottomData, vector<TopBottomData>, TopBottomComparison > TopBottomQueue;
    TopBottomQueue wienerIndexTop100(TopBottomComparison(true)); // The top 100 queues are inverse priority queues
    TopBottomQueue diameterTop100(TopBottomComparison(true));
    TopBottomQueue wienerIndexBottom100; // The bottom 100 queues are normal priority queues
    TopBottomQueue diameterBottom100;
    const unsigned MAX_QUEUE_SIZE = 100;

    //
    // For each value of N in the designated range...
    //
    for (unsigned nCurrent = nStart; nCurrent <= nFinish; nCurrent += 2) {

        // Skip if N is 22, as there are no fullerene graphs with 22 vertices
        if (nCurrent == 22) {
            continue;
        }

        ostringstream convert;
        convert << nCurrent;
        string nString = convert.str();
        if (nCurrent >= 132) {
            nString += "_ipr";
        }

        // File processing timers
        clock_t fileStart;
        clock_t fileEnd;
        cout << "Starting to process file Fullerenes_" << nCurrent << "..." << endl;
        fileStart = clock();

        // Open the corresponding input file
        string inputFileName = "../data/in/Fullerenes_" + nString;
        ifstream inputFile(inputFileName.c_str(), ifstream::binary);
        if (!inputFile) {
            // Or log an error and try to continue if unable to open
            errorLog("Unable to open file: " + inputFileName);
            continue;
        }

        // Create an output file (.csv format)...
        string outputFileName = "../data/out/wiener-diameter/Wiener_diameter_" + nString + ".csv";
        ofstream outputFile(outputFileName.c_str());
        if (!outputFile) {
            // Or log an error and try to continue if unable to create.
            errorLog("Unable to create file: " + outputFileName);
            inputFile.close();
            continue;
        }

        // And print the headers of the data we will write.
        outputFile << "isomer_id,wiener_index,diameter" << endl;

        // Read the first 15 characters on the file and check if it
        // really is planar code
        char* planarCodeHeader = new char[15];
        inputFile.get(planarCodeHeader, 16); // one more than buffer size for get to work
        if (strcmp(planarCodeHeader, ">>planar_code<<") != 0) {
            // Or log an error and try to continue if the header is wrong
            errorLog("File is missing planar code header: " + inputFileName + "    Header: " + planarCodeHeader);
            inputFile.close();
            outputFile.close();
            delete[] planarCodeHeader;
            continue;
        }

        // Create a distance matrix that will be used (further below)
        // by all the graphs in the same file (they all have the same size, N)
        unsigned** distances = new unsigned*[nCurrent];
        for (unsigned i = 0; i < nCurrent; ++i) {
            distances[i] = new unsigned[nCurrent];
        }

        // Now there can be any number of graphs encoded on the file, so we
        // give a unique ID to each one, starting at 1.
        unsigned isomerNumber = 1;

        // Each one of these potential graphs starts with a byte that
        // (interpreted as an unsigned integer) represents the number
        // of vertices (n). So while there is a byte to read in the file...
        char byte;
        while (inputFile.get(byte)) {

            // Start graph reading timer
            /*clock_t graphStart;
            clock_t graphEnd;
            cout << "Reading Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            // Instead of creating the whole graph, we'll just use the previously
            // created distance matrix and feed it into the wienerIndexAndDiameter
            // function down below (it just does a Floyd-Warshall and returns the
            // desired values).
            // Remembering to reset it to default values before use.
            unsigned numberOfVertices = (unsigned)((unsigned char)byte);
            for (unsigned i = 0; i < numberOfVertices; ++i) {
                for (unsigned j = 0; j < numberOfVertices; ++j) {
                    distances[i][j] = MAX_UNSIGNED;
                }
                distances[i][i] = 0;
            }

            // The next part of the file is divided into n sections. Each
            // section 1 <= v <= n contains the neighbours of vertex v,
            // presented in clockwise order, followed by a zero byte.
            for (unsigned v = 1; v <= numberOfVertices; ++v) {

                // Let w be the current neighbour of v being considered
                inputFile.get(byte);
                unsigned w = (unsigned)((unsigned char)byte);
                while (w != 0) {
                    distances[v - 1][w - 1] = 1;
                    inputFile.get(byte);
                    w = (unsigned)((unsigned char)byte);
                }
            }

            // Finished reading graph data.
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Start calculating
            /*cout << "Calculating Wiener Index and Diameter of Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            unsigned diameter = 0;
            unsigned wienerIndex = wienerIndexAndDiameter(distances, numberOfVertices, diameter);

            // Calulation complete
            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds." 
                 << endl;*/

            // Update top/bottom 100 queues
            if (wienerIndexTop100.size() < MAX_QUEUE_SIZE || wienerIndex > wienerIndexTop100.top().value) {
                if (wienerIndexTop100.size() >= MAX_QUEUE_SIZE) {
                    wienerIndexTop100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = wienerIndex;
                wienerIndexTop100.push(tpd);
            }
            if (diameterTop100.size() < MAX_QUEUE_SIZE || diameter > diameterTop100.top().value) {
                if (diameterTop100.size() >= MAX_QUEUE_SIZE) {
                    diameterTop100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = diameter;
                diameterTop100.push(tpd);
            }
            if (wienerIndexBottom100.size() < MAX_QUEUE_SIZE || wienerIndex < wienerIndexBottom100.top().value) {
                if (wienerIndexBottom100.size() >= MAX_QUEUE_SIZE) {
                    wienerIndexBottom100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = wienerIndex;
                wienerIndexBottom100.push(tpd);
            }
            if (diameterBottom100.size() < MAX_QUEUE_SIZE || diameter < diameterBottom100.top().value) {
                if (diameterBottom100.size() >= MAX_QUEUE_SIZE) {
                    diameterBottom100.pop();
                }
                TopBottomData tpd;
                tpd.isomerID = isomerNumber;
                tpd.value    = diameter;
                diameterBottom100.push(tpd);
            }

            // Write results to file
            /*cout << "Writing results for Graph #" << isomerNumber << "... ";
            graphStart = clock();*/

            outputFile << isomerNumber << "," << wienerIndex << "," << diameter << endl;

            /*graphEnd = clock() ;
            cout << "Done. Total time: "
                 << ( (double) ( graphEnd - graphStart ) ) / CLOCKS_PER_SEC 
                 << " seconds."
                 << endl;*/

            // Increment the counter to be used as an ID for the next graph (if any).
            ++isomerNumber;
        }

        //
        // All the graphs in the file were processed, close it and destroy the auxiliary arrays.
        //
        inputFile.close();
        outputFile.close();
        delete[] planarCodeHeader;
        for (unsigned i = 0; i < nCurrent; ++i) {
            delete[] distances[i];
        }
        delete[] distances;

        //
        // Write to file the Top/Bottom 100 data corresponding to the graphs in this file
        //

        // Get how many numbers are actually on the queues and create vectors with them
        unsigned  countWienerTop      = wienerIndexTop100.size();
        unsigned  countDiameterTop    = diameterTop100.size();
        unsigned  countWienerBottom   = wienerIndexBottom100.size();
        unsigned  countDiameterBottom = diameterBottom100.size();
        vector<unsigned> wienerIndexTop(countWienerTop, 0);
        vector<unsigned> diameterTop(countDiameterTop, 0);
        vector<unsigned> wienerIndexBottom(countWienerBottom, 0);
        vector<unsigned> diameterBottom(countDiameterBottom, 0);

        // Note that the pop method for the Top queues would return
        // numbers smallest > largest, and we want them largest > smallest
        for (unsigned i = countWienerTop; i > 0; --i) {
            wienerIndexTop[i - 1]    = wienerIndexTop100.top().isomerID;
            wienerIndexTop100.pop();
        }
        for (unsigned i = countDiameterTop; i > 0; --i) {
            diameterTop[i - 1]       = diameterTop100.top().isomerID;
            diameterTop100.pop();
        }
        // For the Bottom queues, the opposite is true, but the solution is the same
        for (unsigned i = countWienerBottom; i > 0; --i) {
            wienerIndexBottom[i - 1] = wienerIndexBottom100.top().isomerID;
            wienerIndexBottom100.pop();
        }
        for (unsigned i = countDiameterBottom; i > 0; --i) {
            diameterBottom[i - 1]    = diameterBottom100.top().isomerID;
            diameterBottom100.pop();
        }

        // Now write the contents of the arrays to the correct columns for the Wiener Index...
        wienerIndexTop100Bottom100File << nCurrent << ",\"";
        for (unsigned i = 0; i < countWienerTop; ++i) {
            wienerIndexTop100Bottom100File << wienerIndexTop[i] << endl;
        }
        wienerIndexTop100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countWienerBottom; ++i) {
            wienerIndexTop100Bottom100File << wienerIndexBottom[i] << endl;
        }
        wienerIndexTop100Bottom100File << "\",\"";

        // And for the Diameter
        diameterTop100Bottom100File << nCurrent << ",\"";
        for (unsigned i = 0; i < countDiameterTop; ++i) {
            diameterTop100Bottom100File << diameterTop[i] << endl;
        }
        diameterTop100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countDiameterBottom; ++i) {
            diameterTop100Bottom100File << diameterBottom[i] << endl;
        }
        diameterTop100Bottom100File << "\",\"";

        // Now sort the contents of the 4 arrays...
        sort(wienerIndexTop.begin()   , wienerIndexTop.end());
        sort(diameterTop.begin()      , diameterTop.end());
        sort(wienerIndexBottom.begin(), wienerIndexBottom.end());
        sort(diameterBottom.begin()   , diameterBottom.end());

        // And print the sorted data
        for (unsigned i = 0; i < countWienerTop; ++i) {
            wienerIndexTop100Bottom100File << wienerIndexTop[i] << endl;
        }
        wienerIndexTop100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countWienerBottom; ++i) {
            wienerIndexTop100Bottom100File << wienerIndexBottom[i] << endl;
        }
        wienerIndexTop100Bottom100File << "\"" << endl;

        for (unsigned i = 0; i < countDiameterTop; ++i) {
            diameterTop100Bottom100File << diameterTop[i] << endl;
        }
        diameterTop100Bottom100File << "\",\"";
        for (unsigned i = 0; i < countDiameterBottom; ++i) {
            diameterTop100Bottom100File << diameterBottom[i] << endl;
        }
        diameterTop100Bottom100File << "\"" << endl;


        // Now reset the priority_queue objects to be used on the next file
        wienerIndexTop100    = TopBottomQueue(TopBottomComparison(true)) ;
        diameterTop100       = TopBottomQueue(TopBottomComparison(true)) ;
        wienerIndexBottom100 = TopBottomQueue() ;
        diameterBottom100    = TopBottomQueue() ;

        // Check time
        fileEnd = clock() ;
        cout << "Total time to process file: "
             << ( (double) ( fileEnd - fileStart ) ) / CLOCKS_PER_SEC 
             << " seconds." 
             << endl << endl;

    }

    // Close files
    wienerIndexTop100Bottom100File.close();
    diameterTop100Bottom100File.close();
  
    cout << "Finished.";
    if (errorDetected) {
        cout << " There were some errors, please review error log.";
    } else {
        cout << " Everything OK.";
    }
    cout << endl << endl;
  
    return EXIT_SUCCESS;

}
