/** 
 * \file WienerDiameter.h
 *  
 * \brief Implemention of a simple algorithm to calculate the Wiener
 * Index and Diameter of a fullerene graph.
 *
 * \author
 * Thiago Henrique de Araujo Lemos \n
 * Universidade Federal do Paraná, \n
 * Departamento de Informatica, \n
 * thalemos at inf (dot) ufpr (dot) br
 *
 * \version 1.0
 * \date March 2014
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */
 
#ifndef _WIENER_DIAMETER_H_
#define _WIENER_DIAMETER_H_


//#include <cassert>                          // assert
#include <limits>                           // numeric_limits::max

using namespace std;

const unsigned MAX_UNSIGNED = numeric_limits<unsigned int>::max();

// ------------------------------------------------------------------
//
// Implementation starts here.
//
// ------------------------------------------------------------------

/**
 * \fn unsigned wienerIndexAndDiameter(unsigned** distances , unsigned n, unsigned& diameter)
 *
 * \brief Implementation of a simple algorithm to calculate the wiener
 * index and diameter of a given fullerene graph (in the form of a distance matrix).
 *
 * \param distances Distance matrix of the graph.
 * \param n         Number of vertices
 * \param diameter  Will contain the diameter of the graph at the end of the call.
 *
 * \return The wiener index of the graph.
 */
unsigned wienerIndexAndDiameter(unsigned** distances, unsigned n, unsigned& diameter) {

    // Simple Floyd-Warshall algorithm

    for (unsigned k = 0; k < n; ++k) {
        for (unsigned i = 0; i < n; ++i) {
            for (unsigned j = 0; j < n; ++j) {
                if (distances[i][k] != MAX_UNSIGNED && distances[k][j] != MAX_UNSIGNED && distances[i][j] > distances[i][k] + distances[k][j]) {
                    distances[i][j] = distances[i][k] + distances[k][j];
                }
            }
        }
    }

    // Now retrieve the values from the matrix
    unsigned wienerIndex = 0;
    diameter = 0;
    for (unsigned i = 0; i < n; ++i) {
        for (unsigned j = 0; j < n; ++j) {
            if (i < j) {
                wienerIndex += distances[i][j];
            }
            if (diameter < distances[i][j]) {
                diameter = distances[i][j];
            }
        }
    }

    return wienerIndex;
    
}

#endif // _WIENER_DIAMETER_H_
